function flow_signals = read_flow_signals(file_name,frame_rate)


% Non provided input arguments are initially assigned an empty
% matrix to indicate they should get their default values
if (nargin < 2), frame_rate = []; end;

% If the input argument 'frame_rate' is not provided, it defaults
% to the NTSC frame rate (29.97 fps). However, if this value
% is later found in the loaded optical flow file, or if it can
% be derived from any variable in that file (e.g., a time vector
% array), that value will override the rate set here.
if isempty(frame_rate), frame_rate = 29.97; end;

try
	
	% First, we try to read the file using Matlab's function load().
	% This works with the old optical flow signal files, which are
	% saved as Matlab mat files.
	flow_signals = load(file_name);
	
	% Check if the loaded file is indeed a flow signals file
	check_file_contents(flow_signals);
	
	% Get the initial and final frames from the file name
	frame_borders = frame_borders_from_filename(file_name);
	
	% Add a time vector to the structure 'flow_signals'
	flow_signals = create_time_vector(flow_signals,frame_borders,frame_rate);
	
catch
	
	% If reading the file as a Matlab mat file doesn't work, we try
	% reading it as an hdf file. There are two functions for reading
	% HDF files: read_flow_signals_v01() and read_flow_signals_v02().
	% We have to choose between them depending on the version of Matlab
	% we're running on. In order to avoid checking the Matlab version,
	% we simply call the first function and, if that produces an error,
	% we call the second one.
	try
		flow_signals = read_flow_signals_v01(file_name);
	catch
		flow_signals = read_flow_signals_v02(file_name);
	end
	
	% Check if the loaded file is indeed a flow signals file
	check_file_contents(flow_signals);
	
end

% Compute a new magnitude signal 'sum_mag' which is simply the magnitude
% of the resultant vector
flow_signals = compute_magnitudes(flow_signals);

% Check the signal lengths to make sure all signals have the same length
flow_signals = check_signal_lengths(flow_signals);

% --------------------------------------------------------------------------------------------- %

function check_file_contents(flow_signals)

% The names of the variables loaded from file
field_names = fieldnames(flow_signals);

% If there isn't a variable 'sums' in the loaded file, this isn't
% a flow signals file and thus there is nothing to do
if ~ismember('sums',field_names)
	error('This does not seem to be a flow signals file. Exiting...');
end

% --------------------------------------------------------------------------------------------- %

function frame_borders = frame_borders_from_filename(file_name)

% This function extracts the initial and final frames of the
% optical flow analysis from the name of the optical flow file

pattern = '_frames_(\d+)_(\d+)';
aux = regexpi(file_name,pattern,'tokens','once');
frame_borders = cellfun(@str2num,aux);

% --------------------------------------------------------------------------------------------- %

function flow_signals = create_time_vector(flow_signals,frame_borders,frame_rate)

% Add a time vector to the structure 'flow_signals' with a sampling
% rate of 1 Hz
flow_signals.time_vector = [frame_borders(1):frame_borders(2)]'/frame_rate;

% --------------------------------------------------------------------------------------------- %

function flow_signals = check_signal_lengths(flow_signals)

% This function ensures that all signals in the 'flow_signals' structure
% have the same length

% The region names
region_names = fieldnames(flow_signals.sums);

% The number of samples in the signals of each region of interest
for m=1:length(region_names)
	n_samples{m} = structfun(@length,flow_signals.sums.(region_names{m}));
end

% The number of samples in the shortest signal (note, however, that all signals
% should always have the same number of samples)
n_samples = cell2mat(n_samples);
n_samples = min(n_samples(:));

% Loop over regions of interest
for m=1:length(region_names)
	
	% The name of the current region
	region_name = region_names{m};
	
	% The signal names for the current region
	signal_names = fieldnames(flow_signals.sums.(region_name));
	
	for n=1:length(signal_names)
		
		% The name of the current signal
		signal_name = signal_names{n};
		
		% Take only the first 'n_samples' samples of the signal
		flow_signals.sums.(region_name).(signal_name) = flow_signals.sums.(region_name).(signal_name)(1:n_samples);
		
	end
	
end

% Make sure the time vector has the same number of samples that
% the motion signals themselves
flow_signals.time_vector = flow_signals.time_vector(1:n_samples);

% --------------------------------------------------------------------------------------------- %

function flow_signals = compute_magnitudes(flow_signals)

% The names of the regions of interest
region_names = fieldnames(flow_signals.sums);

% The number of regions of interest
n_regions = length(region_names);

% Loop over regions of interest
for m=1:n_regions
	
	% The name of the current region of interest
	region_name = region_names{m};
	
	% If the fied 'mag_mag' already exists, there's nothing to do
	% if isfield(flow_signals(m).signals,'mag_mag'), continue; end;
	if isfield(flow_signals.sums.(region_name),'sum_mag'), continue; end;
	
	% The current 'mag' signal contains the sum of the magnitudes of all
	% individual flow vectors. We now compute a new signal 'sum_mag' as
	% simply the magnitude of the resultant vector
	flow_signals.sums.(region_name).sum_mag = ...
		sqrt(flow_signals.sums.(region_name).x.^2 + flow_signals.sums.(region_name).y.^2);
	
end

% --------------------------------------------------------------------------------------------- %
