# -*- coding: utf-8 -*-

# This is flow_lib, a Python module for optical flow analysis based on OpenCV.
# Author: Adriano Vilela Barbosa (adriano.vilela@gmail.com)
# Copyright 2006-2014 Adriano Vilela Barbosa


# ----------------------------------------------------------------------------------------------- #

# Changelog:

# - v06 (2017/01/06) - code ported from PyTables 2.x to 3.x (see http://www.pytables.org/release-notes/RELEASE_NOTES_v3.0.x.html and http://www.pytables.org/MIGRATING_TO_3.x.html for details)

# ----------------------------------------------------------------------------------------------- #


# Force the division of two integers to float
# from __future__ import division

import os
import pdb
import sys
import time
import glob
import struct
import numpy
# import Image
from PIL import Image
import warnings
import tables
# import pyffmpeg
import optparse
import notify.all as notify

# Set Matplotlib's backend (Qt4, Qt5, etc)
import matplotlib
# matplotlib.use('Qt4Agg')
# matplotlib.use('Qt5Agg')

import matplotlib.pyplot as plt
import xlwt, xlrd

# OpenCV imports (*new* Python interface)
import cv

#CLASS IMPLEMENTATION:
#
# CameraReader(object):				DONE
# MovieReaderOpencv(object):			DONE
# MovieReaderPyffmpeg(object):		DONE
# FlowAnalysis(object):				DONE
# FlowFileBinary(object):				DONE
# FlowFileHDF(object):					IN PROGRESS
# FlowApplication(object):			DONE


# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

#Classes in this module:
#
#- CameraReader        (class for reading a video stream from camera)
#- MovieReaderOpencv   (class for reading movie files - OpenCV backend)
#- MovieReaderPyffmpeg (class for reading movie files - pyffmpeg backend)
#- FlowAnalysis        (class for performing optical flow analysis)
#- FlowFileBinary      (class for both reading and writing binary (not hdf5) optical flow files)
#
#TODO: We need at least another two classes:
#
#- FlowRegions (class defining regions of interest and disinterest in the optical flow file)
#- FlowReader (class for reading an optical flow file, while processing rois and rods passed in the FlowRegions object)

#- We have to implement a new class for reading/writing hdf5 files.
#- It may be a good idea to split the class FlowFileBinary() into a reader
#  (FlowFileBinaryReader) and a writer (FlowFileBinaryWriter) class

# In order to implement the HDF5 reader/writer, I think we have to look at the files
# optical_flow_mac_sid_h5py.py and opencv_lib_3.py. Update: that's right; these are
# the two files we should be looking at


# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

class CameraReader(object):

	# IMPORTANT: THIS CLASS IS STILL USING THE OLD PYTHON INTERFACE. WE HAVE TO UPDATE IT
	# TO USE THE NEW INTERFACE, WHERE EVERYTHING IS IMPORTED UNDER THE MODULE CV

	def __init__(self,camera_id=-1):

		# The class' attributes
		self.camera_id    = None
		self.input_video  = None
		self.width        = None
		self.height       = None
		self.frame_rate   = None
		self.frame        = None
		self.frame_number = -1

		# The name of the input video file
		# self.input_file = input_file
		self.camera_id = camera_id

		# Open capture
		self.open_capture()

# ----------------------------------------------------------------------------------------------- #

	def open_capture(self):

		# Initializes capture source. After the allocated structure is not
		# used any more it should be released by cvReleaseCapture function.
		# input_video = highgui.cvCaptureFromFile(input_file)
		#input_video = highgui.cvCreateFileCapture(input_file)
		#input_video = highgui.cvCreateCameraCapture(-1)
		#self.input_video = highgui.cvCreateCameraCapture(-1)
		#self.input_video = highgui.cvCreateCameraCapture(self.camera_id)
		self.input_video = cv.CreateCameraCapture(self.camera_id)

		# Checks if the video file opening was successful
		if not self.input_video:
			print "Error opening input video file. Quitting..."
			sys.exit(-1)

		# If we got to this point, that means the input video has been opened successfully
		# and, therefore, we can extract some pieces of information from it
#		self.height = int(highgui.cvGetCaptureProperty(self.input_video,highgui.CV_CAP_PROP_FRAME_HEIGHT))
#		self.width  = int(highgui.cvGetCaptureProperty(self.input_video,highgui.CV_CAP_PROP_FRAME_WIDTH))
#		# self.frame_rate = highgui.cvGetCaptureProperty(self.input_video,highgui.CV_CAP_PROP_FPS)
		self.height = int(cv.GetCaptureProperty(self.input_video,cv.CV_CAP_PROP_FRAME_HEIGHT))
		self.width  = int(cv.GetCaptureProperty(self.input_video,cv.CV_CAP_PROP_FRAME_WIDTH))
		# self.frame_rate = cv.GetCaptureProperty(self.input_video,cv.CV_CAP_PROP_FPS)

# ----------------------------------------------------------------------------------------------- #

	def read_frame(self,frame_number):

		#Seeking is not supported when reading video from a camera. Nonetheless, we define
		#this method (which simply reads the next frame) so that the CameraReader class has
		#exactly the same protocol as the MovieReaderOpencv and MovieReaderPyffmpeg classes.
		print "Seeking is not supported when reading video from camera. Reading the next frame..."

		# At this point, we are at the correct position in the video and therefore
		# we just read the next frame
		self.read_next_frame()

# ----------------------------------------------------------------------------------------------- #

	def read_next_frame(self):

		# Read the first video frame. On Linux, video properties can only be queried
		# after the first frame has been read. This seems to be a problem with ffmpeg,
		# which is used by OpenCV on Linux. It is important to note that cvQueryFrame()
		# always returns a pointer to the *same* memory location. Therefore, successive
		# calls frame1 = cvQueryFrame(); frame2 = cvQueryFrame(); frame3 = cvQueryFrame();
		# will result in frame1 == frame2 == frame3. The solution to this is to make a
		# copy of the cvQueryFrame output. Furthermore, we should note that it is not
		# necessary to pre-allocate memory for the image returned by cvQueryFrame(),
		# nor should this image be released or modified by the user (according to
		# the OpenCV wiki page (http://opencvlibrary.sourceforge.net/HighGui).
		# frame = cvQueryFrame(input_video);
		# frame = highgui.cvQueryFrame(input_video)
		# self.frame = highgui.cvQueryFrame(self.input_video)
		self.frame = cv.QueryFrame(self.input_video)

# ----------------------------------------------------------------------------------------------- #

	def close_capture(self):

		pass

		# Important: the function cvReleaseCapture() is not present in the new Python
		# interface of OpenCV 2.1, which makes me think that we don't need to manually
		# release capture sources anymore. This makes this method unnecessary.

		# Release the capture source
		# cv.cvReleaseCapture(self.input_video)

# ----------------------------------------------------------------------------------------------- #

	def __del__(self):

		# Important: Apparently, we should avoid using destructors
		# in Python and use contexts ("with" statements) instead

		# This is the destructor function

		# Close the optical flow file
		self.close_capture()

# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

class MovieReaderOpencv(object):

	# Signals emitted by this class
	video_frame_read = notify.Signal()

# ----------------------------------------------------------------------------------------------- #


	def __init__(self,input_file):

		# The class' attributes
		self.input_file   = None
		self.input_video  = None
		self.width        = None
		self.height       = None
		self.n_frames     = None
		self.frame_rate   = None
		self.frame        = None
		self.frame_number = -1

		# The name of the input video file
		self.input_file = input_file

		# Open capture
		self.open_capture()

# ----------------------------------------------------------------------------------------------- #

	def open_capture(self):

		# Initializes capture source. After the allocated structure is not
		# used any more it should be released by cvReleaseCapture function.
		# input_video = highgui.cvCaptureFromFile(input_file)
		# input_video = highgui.cvCreateFileCapture(input_file)
		# self.input_video = highgui.cvCreateFileCapture(self.input_file)
		self.input_video = cv.CreateFileCapture(self.input_file)

		# Checks if the video file opening was successful
		if not self.input_video:
			print "Error opening input video file. Quitting..."
			sys.exit(-1)

		# If we got to this point, that means the input video has been opened successfully
		# and, therefore, we can extract some pieces of information from it
		# self.height = int(highgui.cvGetCaptureProperty(self.input_video,highgui.CV_CAP_PROP_FRAME_HEIGHT))
		# self.width  = int(highgui.cvGetCaptureProperty(self.input_video,highgui.CV_CAP_PROP_FRAME_WIDTH))
		# self.frame_rate = highgui.cvGetCaptureProperty(self.input_video,highgui.CV_CAP_PROP_FPS)
		self.height = int(cv.GetCaptureProperty(self.input_video,cv.CV_CAP_PROP_FRAME_HEIGHT))
		self.width  = int(cv.GetCaptureProperty(self.input_video,cv.CV_CAP_PROP_FRAME_WIDTH))
		self.frame_rate = cv.GetCaptureProperty(self.input_video,cv.CV_CAP_PROP_FPS)
		self.n_frames = int(cv.GetCaptureProperty(self.input_video,cv.CV_CAP_PROP_FRAME_COUNT))

# ----------------------------------------------------------------------------------------------- #

	## IMPORTANT: THIS METHOD WAS AN ATTEMPT TO WORK AROUND THE PROBLEMS WE HAVE WHEN TRYING TO SEEK
	## TO A PARTICULAR FRAME WHEN USING OPENCV'S cv.SetCaptureProperty() FUNCTION. THE WORKAROUND
	## WORKED FINE WITH DEBIAN'S OFFICIAL LIBAV PACKAGES, BUT NOT WITH MARILLAT'S LIBAV PACKAGES
	## AVAILABLE AT WWW.DEBIAN-MULTIMEDIA.ORG. IN THE CASE OF THE LATTER LATTER PACKAGES, THE
	## FUNCTION cv.SetCaptureProperty() WAS COMPLETELY BROKEN FOR THE H264 FILES WE TESTED.
	## BASICALLY, AFTER A CALL TO cv.QueryFrame(), TRYING TO USE cv.SetCaptureProperty()
	## TO SEEK TO A GIVEN FRAME HAD NO EFFECT WHATSOEVER. SO, I'LL JUST FORGET ABOUT IT
	## UNTIL THE NEXT VERSION OF OPENCV (2.4) IS RELEASED.


	#def read_frame(self,frame_number):

		## If 'frame_number' is the next frame (i.e., self.frame_number + 1), we simply
		## call the method read_next_frame(), as no seeking is necessary in this case.
		## Otherwise, we seek to the appropriate frame before calling read_next_frame().
		#if (frame_number != self.frame_number + 1):

			## Go to frame 'frame_number'. Very important: this changes the current position
			## in the video so that a call to cv.QueryFrame() causes the frame 'frame_number'
			## to be read. However, this does not work with H264 videos (at least the ones
			## we tested; it may work on systems other than Linux or maybe with some other
			## combination of H264 encoding parameters, such as frame re-ordering, etc).
			## In normal (non-problematic) videos, a call to cv.SetCaptureProperty() will
			## point the current frame to 'frame_number' so that a call to cv.QueryFrame()
			## will result in that frame being read and the current frame being updated to
			## 'frame_number' + 1. However, in the problematic videos we tried, a call to
			## cv.SetCaptureProperty() updates the current frame to 'frame_number' only
			## AFTER a call to cv.QueryFrame(). This call to cv.QueryFrame() will cause
			## whatever the current frame is to be read and then be updated to point to
			## 'frame_number'. A new call to cv.QueryFrame() will then cause the frame
			## 'frame_number' to be read and the current frame to be updated to point
			## to 'frame_number' + 1. This is summarized below:
			## Normal videos:
			## - current frame is X
			## - cv.SetCaptureProperty(frame_number)
			## - cv.QueryFrame() -> reads frame 'frame_number'
			## - current frame is now 'frame_number' + 1
			## Problematic videos:
			## - current frame is X
			## - cv.SetCaptureProperty(frame_number) -> this has no effect at this point
			## - cv.QueryFrame() -> reads frame 'X'
			## - current frame is now 'frame_number'
			## - cv.QueryFrame() -> reads frame 'frame_number'
			## - current frame is now 'frame_number' + 1
			#cv.SetCaptureProperty(self.input_video,cv.CV_CAP_PROP_POS_FRAMES,frame_number)


		## Our initial idea for the read_frame() method was to simply point the current
		## frame to 'frame_number' and then call the method read_next_frame() to read
		## that frame. That can be done (like in the loop commented out below) and works
		## fine. However, in the case of the problematic H264 videos, the loop below runs
		## twice (the whole point of having the loop in the first place is to work around
		## these videos) and, therefore, the method read_next_frame() is called twice.
		## This method emits the signal 'video_frame_read' every time it is executed
		## and therefore the signal would be emitted twice. However, in the case of the
		## H264 videos, the signal should be emitted only once, after the correct frame
		## is read. In order to do that, we have to do avoid calling read_next_frame()
		## and do the reading ourselves from here.

##		# Keep calling read_next_frame() until the correct frame is read. This loop
##		# should run only once in the case of normal (non-problematic) videos and
##		# twice in the case of problematic videos (see details above).
##		while (self.frame_number != frame_number):
##			self.read_next_frame()

		## Note: if the problem with the H264 videos is ever solved in OpenCV, we can
		## go back to the read_next_frame() approach by uncommenting the loop above
		## and getting rid of everything below.

		## Keep calling cv.QueryFrame() until the correct frame is read. This loop
		## should run only once in the case of normal (non-problematic) videos and
		## twice in the case of problematic videos (see details above).
		#while (self.frame_number != frame_number):
			#self.frame = cv.QueryFrame(self.input_video)
			#self.frame_number = int(cv.GetCaptureProperty(self.input_video,cv.CV_CAP_PROP_POS_FRAMES)) - 1

		## Emit the signal 'video_frame_read'
		#self.video_frame_read(self.frame,self.frame_number)

# ----------------------------------------------------------------------------------------------- #

	def read_frame(self,frame_number):

		# If 'frame_number' is the next frame (i.e., self.frame_number + 1), we simply
		# call the method read_next_frame(), as no seeking is necessary in this case.
		# Otherwise, we seek to the appropriate frame before calling read_next_frame().
		if (frame_number != self.frame_number + 1):

			# Go to frame 'frame_number'
			cv.SetCaptureProperty(self.input_video,cv.CV_CAP_PROP_POS_FRAMES,frame_number)

			# Update the frame number. This is necessary because the method read_next_frame()
			# increments the frame number by one
			self.frame_number = frame_number - 1

		# At this point, we are at the correct position in the video and therefore
		# we just read the next frame
		self.read_next_frame()

		# Emit the signal 'video_frame_read'
		self.video_frame_read(self.frame,self.frame_number)

# ----------------------------------------------------------------------------------------------- #

	def read_next_frame(self):

		# Read the next video frame
		self.frame = cv.QueryFrame(self.input_video)

		# Update the frame number
		self.frame_number += 1

		# Emit the signal 'video_frame_read'
		self.video_frame_read(self.frame,self.frame_number)

# ----------------------------------------------------------------------------------------------- #

	def close_capture(self):

		pass

		# Important: the function cvReleaseCapture() is not present in the new Python
		# interface of OpenCV 2.1, which makes me think that we don't need to manually
		# release capture sources anymore. This makes this method unnecessary.

		# Release the capture source
		# cv.cvReleaseCapture(self.input_video)

# ----------------------------------------------------------------------------------------------- #

	def __del__(self):

		# Important: Apparently, we should avoid using destructors
		# in Python and use contexts ("with" statements) instead

		# This is the destructor function

		# Close the optical flow file
		self.close_capture()

# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

#class MovieReaderPyffmpeg(object):
#
#	# These dictionaries are used when opening the FFMpegReader object
#	TS_VIDEO_RGB24 = {'video1':(0, -1, {'pixel_format':pyffmpeg.PixelFormats.RGB24})}
#	TS_VIDEO_BGR24 = {'video1':(0, -1, {'pixel_format':pyffmpeg.PixelFormats.BGR24})}
#	TS_VIDEO_GRAY8 = {'video1':(0, -1, {'pixel_format':pyffmpeg.PixelFormats.GRAY8})}
#
#	# Signals emitted by this class
#	video_frame_read = notify.Signal()
#
## ----------------------------------------------------------------------------------------------- #
#
#	def __init__(self,input_file):
#
#		# The class' attributes
#		self.input_file   = None
#		self.input_video  = None
#		self.width        = None
#		self.height       = None
#		self.n_frames     = None
#		self.frame_rate   = None
#		self.frame        = None
#		self.frame_number = -1
#
#		# The name of the input video file
#		self.input_file = input_file
#
#		# Open capture
#		self.open_capture()
#
## ----------------------------------------------------------------------------------------------- #
#
#	def open_capture(self):
#
#		# Create the movie reader object
#		media_reader = pyffmpeg.FFMpegReader(0,False)
#
#		# Open the media file
#		media_reader.open(self.input_file,self.TS_VIDEO_RGB24)
#
#		# The video track of interest
#		tracks = media_reader.get_tracks()
#		video_track = tracks[0]
#
#		# Set the observer
#		#video_track.set_observer(self.process_frame)
#		#video_track.set_observer(self.show_image)
#
#		#self.media_reader = media_reader
#		#self.video_track = video_track
#		self.input_video = video_track
#
#		# If we got to this point, that means the input video has been opened successfully
#		# and, therefore, we can extract some pieces of information from it
#		(self.width, self.height) = self.input_video.get_size()
#		self.frame_rate = self.input_video.get_fps()
#		self.n_frames = len(self.input_video)
#
## ----------------------------------------------------------------------------------------------- #
#
#	def read_frame(self,frame_number):
#
#		# If 'frame_number' is the next frame (i.e., self.frame_number + 1), we simply
#		# call the method read_next_frame(), as no seeking is necessary in this case.
#		# Otherwise, we seek to the appropriate frame before calling read_next_frame().
#		if (frame_number != self.frame_number + 1):
#
#			# FIXME: THE CODE BELOW PRODUCES AN ERROR WHEN 'frame_number' IS ZERO.
#			# APPARENTLY, IT'S IMPOSSIBLE TO READ FRAME ZERO WITH PYFFMPEG. WE
#			# SHOULD TEST THIS WITH A NEWER VERSION OF PYFFMPEG AND REPORT A BUG
#			# IN CASE THE PROBLEM IS STILL PRESENT.
#
#			# Go to frame 'frame_number'
#			# Important: in OpenCV, when we seek to frame N (using cv.SetCaptureProperty())
#			# and then read the next frame (using cv.QueryFrame()), we get frame N. In
#			# Pyffmpeg, when we seek to frame N (using input_video.seek_to_frame()) and
#			# then read the next frame (using input_video.get_next_frame()), we get frame
#			# N+1 instead. Because of that, when using Pyffmpeg to read frame N , we have
#			# to seek to frame N-1 (not N) before calling input_video.get_next_frame().
#			# self.input_video.seek_to_frame(frame_number - 1)
#			self.input_video.seek_to_frame(frame_number)
#
#			# Update the frame number. This is necessary because the method read_next_frame()
#			# increment the frame number by one)
#			self.frame_number = frame_number - 1
#
#		# At this point, we are at the correct position in the video and therefore
#		# we just read the next frame
#		self.read_next_frame()
#
## ----------------------------------------------------------------------------------------------- #
#
#	def read_next_frame(self):
#
#		# Read the next video frame. The frame is returned as a numpy array
#		aux_frame = self.input_video.get_next_frame()
#
#		# Convert the read frame from a numpy array to an OpenCV image
#		# TODO: IN THE FUTURE, I THINK WE SHOULD USE NUMPY ARRAYS TO STORE IMAGES
#		# REGARDLESS OF THE VIDEO BACKEND WE'RE USING (EITHER OPENCV OR PYFFMPEG)
#		self.frame = cv.GetImage(cv.fromarray(aux_frame))
#
#		# Increment the frame number
#		self.frame_number += 1
#
#		# Emit the signal 'video_frame_read'
#		self.video_frame_read(self.frame,self.frame_number)
#
## ----------------------------------------------------------------------------------------------- #
#
#	def close_capture(self):
#
#		pass
#
#		# Important: the function cvReleaseCapture() is not present in the new Python
#		# interface of OpenCV 2.1, which makes me think that we don't need to manually
#		# release capture sources anymore. This makes this method unnecessary.
#
#		# Release the capture source
#		# self.input_video.close()
#
## ----------------------------------------------------------------------------------------------- #
#
#	def __del__(self):
#
#		# Important: Apparently, we should avoid using destructors
#		# in Python and use contexts ("with" statements) instead
#
#		# This is the destructor function
#
#		# Close the optical flow file
#		self.close_capture()

# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

class FlowAnalysis(object):


#	# This is used to convert the optical flow frames from float to integer.
#	# Basically, we map the range [-FLOW_RANGE:+FLOW_RANGE] onto the range
#	# [-32768:32767]. Ideally, this value should be computed from the movie
#	# under analysis, but this would require a two-pass approach: in the first
#	# pass we go through all frames in order to compute the flow range; in the
#	# second pass we map each flow frame onto the new range and save it to disk.
#	# For now, we use a statically defined flow range. We use the range [-655:655],
#	# which yields a quantization step of 0.02 for the flow values, which means
#	# the quantization error will be in the range [-0.01:0.01]. This value was
#	# chosen after we ran the optical flow analysis on a wide range of movies
#	# and found no optical flow value above 600 (absolute value).
#	FLOW_RANGE = 655.0



	#----- Some class attributes -----#


	# The optical flow methods
	flow_methods = [
		{'name': 'hs', 'label': 'Horn and Schunk'},
		{'name': 'lk', 'label': 'Lucas-Kanade'}]

#	# The video backends
#	video_backends = [
#		{'name': 'opencv',   'label': 'OpenCV',   'function': MovieReaderOpencv},
#		{'name': 'pyffmpeg', 'label': 'Pyffmpeg', 'function': MovieReaderPyffmpeg}]

	# The video backends
	video_backends = [
		{'name': 'opencv',   'label': 'OpenCV',   'function': MovieReaderOpencv}]


	#----- Signals emitted by this class -----#


	# Signal emitted after an optical flow frame has been computed
	# flow_frame_computed = notify.Signal()
	flow_frame_processed = notify.Signal()

	# Signal emitted after the initial frame is read.
	# initial_frame_read = notify.Signal()
	initial_video_frame_read = notify.Signal()

	# Signal emitted when the optical flow analysis has finished
	# flow_analysis_finished = notify.Signal()
	flow_source_finished = notify.Signal()


# ----------------------------------------------------------------------------------------------- #

	@staticmethod
	def get_default_video_reader_backend():

		# This dictionary maps computer platforms to default video backend names
		# platform_mapping = {"linux": "pyffmpeg", "darwin": "opencv", "win32": "opencv"}
		platform_mapping = {"linux": "opencv", "darwin": "opencv", "win32": "opencv"}

		# The platform name. We need to do this instead of just looking at sys.platform
		# because, on Unix systems, this command returns a string created by concatenating
		# the OS name and the OS version (e.g., linux2). Since the OS version can vary,
		# we'd rather not rely on it to identify the platform
		platform_name, = [platform_name for platform_name in platform_mapping.keys() if platform_name in sys.platform]

		# The video backend name
		video_backend_name = platform_mapping[platform_name]

		# Return the name of the default video backend
		return video_backend_name

# ----------------------------------------------------------------------------------------------- #

	@staticmethod
	def get_default_flow_method():

		return "hs"

# ----------------------------------------------------------------------------------------------- #

	# def __init__(self,input_video,output_file_type="binary",flow_method="hs",flow_depth=16):
	# def __init__(self,input_video,output_file_type="binary",flow_method="hs"):
	# def __init__(self,input_video,flow_method="hs"):
	def __init__(self, input_file_name, flow_method = "hs", video_backend = "opencv"):



#		self.flow_app.open_input_file(self.input_file_name)
#		self.flow_app.set_output_file(self.output_file_name,self.output_file_type['name'])
#		self.flow_app.set_frame_range(self.initial_frame,self.final_frame)
#		# self.flow_app.set_regions()
#		self.flow_app.set_flow_method(self.flow_method['name'])
#		self.flow_app.set_video_reader_backend(self.video_backend['name'])



		# The name of the input file
		self.input_file_name = input_file_name

		# The input video object
		self.input_video = None

		# The resolution of the input video
		self.width  = None
		self.height = None

		# The frame rate of the input video
		self.frame_rate = None

		# The initial and final frames for the analysis
		self.initial_frame = None
		self.final_frame   = None

		# The number of frames in the *optical flow analysis*
		self.n_frames = None

		# The number of the current optical flow frame
		self.frame_number = -1

		# The optical flow method
		self.flow_method = None

		# The video reader backend
		self.video_backend = None


		# This attribute is used to cancel the analysis
		self.flow_analysis_canceled = False

		# Variables holding actual video frames
		self.video_frame = {'original': None, 'screen': None, 'previous': None, 'current': None}

		# Variables holding optical flow frames
		# self.flow_frame = {'analysis': None, 'storage': None}
#		self.flow_frame = {'analysis': None}
#		self.flow_frame['analysis'] = {'x': None, 'y': None, 'mag': None}
#		self.flow_frame['storage'] = {'x': None, 'y': None}
		# self.flow_frame = {'x': None, 'y': None, 'mag': None}
		self.flow_frame = {'x': None, 'y': None}

		# Initialize the flow method to its default value
		self.set_flow_method(flow_method)

		# Initialize the video reader backend to its default value
		self.set_video_reader_backend(video_backend)

		# Open the input file
		self.open_input_file()

		# Allocate the OpenCV images
		self.create_opencv_images()

# ----------------------------------------------------------------------------------------------- #

	def open_input_file(self):

		# Construct the MovieReader object
		self.input_video = self.video_backend['function'](self.input_file_name)

		# Set the input video resolution
		(self.width, self.height) = (self.input_video.width, self.input_video.height)

		# Set the input video's frame rate
		self.frame_rate = self.input_video.frame_rate

		# Set the frame range to its default value (i.e., to the first
		# and last frames in the input video). The frame range can be
		# changed later by calling the 'set_frame_range' method again
		self.set_frame_range()

# ----------------------------------------------------------------------------------------------- #

	def set_flow_method(self, method_name = None):

		# If no file type has been provided, we use the default one
		if (method_name == None):
			method_name = self.get_default_flow_method()

		try:
			# The file type dictionary that matches the provided type name 'type_name'
			(flow_method,) = [flow_method for flow_method in self.flow_methods if flow_method['name'] == method_name]
		except:
			error_message = "Invalid flow method: %s." % method_name
			raise Exception(error_message)

		# A dictionary mapping flow method names onto flow methd functions
		mapping = {'hs': self.optical_flow_hs, 'lk': self.optical_flow_lk}

		# Add a 'function' field with the optical flow method to be used
		# to the 'flow_method' dictionary
		flow_method['function'] = mapping[flow_method['name']]

		# Update the file type dictionary in the FlowAnalysis object
		self.flow_method = flow_method

# ----------------------------------------------------------------------------------------------- #

	def set_video_reader_backend(self, backend_name = None):

		# If no video backend name has been provided, we use the default one
		if (backend_name == None):
			backend_name = FlowAnalysis.get_default_video_reader_backend()

		try:
			# The video backend dictionary that matches the provided backend name 'backend_name'
			(video_backend,) = [backend for backend in self.video_backends if backend['name'] == backend_name]
		except:
			error_message = "Invalid video backend: %s." % backend_name
			raise Exception(error_message)

		# Update the video backend attribute in the object
		self.video_backend = video_backend

# ----------------------------------------------------------------------------------------------- #

	def read_initial_frame(self):

		# Read the frame number of interest
		self.input_video.read_frame(self.initial_frame)







#			# Update the frame number. This is necessary because the method read_next_frame()
#			# increments the frame number by one
#			self.frame_number = self.initial_frame - 1







		# Move frames forward
		self.__move_frames_forward()

		# Emit the 'initial_video_frame_read' signal. FlowFile objects listen to this
		# signal in order to be notified that the initial video frame has been
		# read so that they can update their headers. FlowFile objects store the
		# initial video frame in their headers but they can't do that when the
		# headers are created because the initial video frame is not available
		# at that time.
		# self.initial_frame_read(self.video_frame['current'])
		self.initial_video_frame_read(self.video_frame['current'])

# ----------------------------------------------------------------------------------------------- #

	def read_next_frame(self):

		# Read the next frame number
		self.input_video.read_next_frame()

		# Move frames forward
		self.__move_frames_forward()

# ----------------------------------------------------------------------------------------------- #

	def __move_frames_forward(self):

		# Copy the 3-channel frame just read from the input video to the
		# frames 'original' and 'screen'
		cv.Copy(self.input_video.frame,self.video_frame['original'])
		cv.Copy(self.input_video.frame,self.video_frame['screen'])

		# Copy frame 'current' to frame 'previous' (both 1-channel frames)
		cv.Copy(self.video_frame['current'],self.video_frame['previous'])

#		#Now, we have to convert the image containing the current video frame (which is
#		#probably a color image) into a grayscale image, so that it can be used in the
#		#optical flow routines. We originally did this (in OpenCV 1.1) using the function
#		#highgui.cvConvertImage(), but this function is no longer present in the new Python
#		#interface of OpenCV 2.1. We tried using the function cv.Convert(), but we kept
#		#getting an error when using it. We then decided to use the function cv.CvtColor(),
#		#which seems to be *the* function for color space conversion (highgui.cvConvertImage()
#		#could additionally do a vertical flip of the image when doing the color conversion).
#		#The problem with cv.CvtColor() is that we need to explicitly define the source and
#		#destination color spaces (highgui.cvConvertImage() did this automatically from the
#		#image structures), which I don't like. Looking at the source code examples provided
#		#with OpenCV 2.1, we noticed that conversions from color to grayscale space done with
#		#cv.CvtColor() always use cv.CV_BGR2GRAY as the third argument to the function, and
#		#thus that's what we do here (please note the difference between cv.CV_BGR2GRAY and
#		#cv.CV_RGB2GRAY).
#		cv.CvtColor(self.video_frame['original'],self.video_frame['current'],cv.CV_BGR2GRAY)

		# Copy frame 'original' to frame 'current', converting it from 3 to 1 channel
		cv.CvtColor(self.video_frame['original'],self.video_frame['current'],cv.CV_BGR2GRAY)

# ----------------------------------------------------------------------------------------------- #

	# def define_frame_range(self,initial_frame,final_frame):
	def set_frame_range(self, initial_frame = None, final_frame = None):

		# The number of frames in the input video
		n_frames = self.input_video.n_frames

		# If the initial frame has not been provided, it defaults to the first frame in the video
		if initial_frame is None:
			initial_frame = 0

		# If the final frame has not been provided, it defaults to the last frame in the video
		if final_frame is None:
			final_frame = n_frames - 1

		# Check if the initial frame is within the valid frame range
		# if initial_frame not in range(0, n_frames):
		if not (0 <= initial_frame <= n_frames):
			error_message = "Invalid initial frame. Frame range is [%g %g]." % (initial_frame, n_frames-1)
			raise Exception(error_message)

		# Check if the final frame is within the valid frame range
		# if final_frame not in range(0, n_frames):
		if not (0 <= final_frame <= n_frames):
			error_message = "Invalid final frame. Frame range is [%g %g]." % (initial_frame, n_frames-1)
			raise Exception(error_message)

		# The final frame must be greater than the initial frame
		if (initial_frame >= final_frame):
			raise Exception("The initial frame must be smaller than the final frame.")

		# Update the frame range attributes ('initial_frame', 'final_frame', 'n_frames')
		# in the FlowAnalysis object
		self.initial_frame = initial_frame
		self.final_frame   = final_frame
		self.n_frames      = self.final_frame - self.initial_frame

# ----------------------------------------------------------------------------------------------- #

	def run(self):


		# Read the initial video frame and use it to update the output file header
		self.read_initial_frame()

		# Pre-process the initial frame
		self.pre_process_frame()


		flow_depth = 32
		aux = 2**(flow_depth-1)
		hist_range = [(-aux, aux-1)]
		n_bins = [100]


#		# histogram = cv.CreateHist(n_bins,cv.CV_HIST_ARRAY,hist_range)
#		hist_x = cv.CreateHist(n_bins,cv.CV_HIST_ARRAY,hist_range)
#		hist_y = cv.CreateHist(n_bins,cv.CV_HIST_ARRAY,hist_range)


		# The main loop. This process up to frame 'self.final_frame', inclusive
		while ((self.input_video.frame_number < self.final_frame) and (not self.flow_analysis_canceled)):

			# Print out a message showing the current frame number
			# print "Processing frame %d / %d" % (self.input_video.frame_number,final_frame-1)
			print "Processing frame %d / %d" % (self.input_video.frame_number,self.final_frame-1)

			# Read the next video frame
			# self.read_next_frame()
			try:
				self.read_next_frame()
			except:
				break


			# Pre-process the current frame
			self.pre_process_frame()

			# Computes optical flow using the chosen method
			self.compute_flow_frame()


#			# Update the cumulative histograms
##			cv.CalcHist([self.flow_frame['analysis']['x']],hist_x,True)
##			cv.CalcHist([self.flow_frame['analysis']['y']],hist_y,True)
#			cv.CalcHist([self.flow_frame['x']],hist_x,True)
#			cv.CalcHist([self.flow_frame['y']],hist_y,True)
#
#
#
#		cv.NormalizeHist(hist_x,1.0)
#		cv.NormalizeHist(hist_y,1.0)



#		# Compute the standard deviations
#		# compute_standard_deviation(file_pointer,n_frames,mean_x,mean_y);
#		sigma_x    = compute_standard_deviation(std_array_x)
#		sigma_y    = compute_standard_deviation(std_array_y)
#		sigma_mag  = compute_standard_deviation(std_array_mag)
#

		# TODO: File header fields we still have to update:
		# - n_sigmas
		# - sigma_x
		# - sigma_y
		# - sigma_mag



		# Emit the 'flow_source_finished' signal to indicate that the analysis
		# is finished. Output flow file objects listen to this signal in order
		# to know when the analysis is done so that they can close their files.
		# The output flow file objects need to do two things upon completion of
		# the analysis: i) write the file header to disk, as it has been updated
		# since it was initially created/written to disk; ii) close the file
		# self.flow_analysis_finished()
		self.flow_source_finished()

# ----------------------------------------------------------------------------------------------- #

	def create_opencv_images(self):

		# The frame size, in pixels
		# frame_size = cv.cvSize(self.width,self.height)
		# frame_size = (self.width, self.height)
		frame_size = (self.input_video.width, self.input_video.height)

		# Allocate an image to store the frame read from the input video file
		self.video_frame['original'] = cv.CreateImage(frame_size,cv.IPL_DEPTH_8U,3)

		# Allocate an image to be shown on screen. This is just an editable
		# copy of the variable 'frame'. For now, it is mostly unused, but we
		# could use it for drawing optical flow arrows and showing them on
		# the screen, for example. Maybe we could use the variable 'frame'
		# for this, but I think it is safer to keep them separate.
		# frame_screen = cv.cvCreateImage(frame_size,cv.IPL_DEPTH_8U,3)
		self.video_frame['screen'] = cv.CreateImage(frame_size,cv.IPL_DEPTH_8U,3)

		# Now we know the frame size in pixels and therefore we can allocate
		# memory for the images. First, we allocate one channel, 8-bit images
		# for holding the video frames to be submitted to optical flow. This
		# is the image format OpenCV algorithms actually operate on (mostly).
		# frame_previous = cv.cvCreateImage(frame_size,cv.IPL_DEPTH_8U,1)
		# frame_current  = cv.cvCreateImage(frame_size,cv.IPL_DEPTH_8U,1)
		self.video_frame['previous'] = cv.CreateImage(frame_size,cv.IPL_DEPTH_8U,1)
		self.video_frame['current']  = cv.CreateImage(frame_size,cv.IPL_DEPTH_8U,1)

		# Allocate "images" for holding the optical flow results
		# flow_x = cv.cvCreateImage(frame_size,cv.IPL_DEPTH_32F,1)
		# flow_y = cv.cvCreateImage(frame_size,cv.IPL_DEPTH_32F,1)
		# flow_mag = cv.cvCreateImage(frame_size,cv.IPL_DEPTH_32F,1)
		# self.flow_frame['analysis']['x']   = cv.CreateImage(frame_size,cv.IPL_DEPTH_32F,1)
		# self.flow_frame['analysis']['y']   = cv.CreateImage(frame_size,cv.IPL_DEPTH_32F,1)
		# self.flow_frame['analysis']['mag'] = cv.CreateImage(frame_size,cv.IPL_DEPTH_32F,1)
		self.flow_frame['x']   = cv.CreateImage(frame_size,cv.IPL_DEPTH_32F,1)
		self.flow_frame['y']   = cv.CreateImage(frame_size,cv.IPL_DEPTH_32F,1)
		# self.flow_frame['mag'] = cv.CreateImage(frame_size,cv.IPL_DEPTH_32F,1)

#		# Create the variables 'flow_x_file' and 'flow_y_file' containing the optical
#		# flow values to be saved to disk, stored as either 16 or 32 bit integers. The
#		# variables 'flow_x' and 'flow_y' contain the optical flow values to be used
#		# in all OpenCV computations, stored as 32-bit floats. If FLOW_DEPTH == 16,
#		# 'flow_x_file' and 'flow_y_file' are created as new OpenCV images with a
#		# 16-bit depth. If FLOW_DEPTH == 32, 'flow_x_file' and 'flow_y_file' are
#		# created as pointers to the variables 'flow_x' and 'flow_y', respectively.
#		# Also, define the scale factor by which we have to multiply the float flow
#		# frame to map it onto an integer flow frame. This value is passed to the
#		# OpenCV function cv.cvConvertScale()
#		if (self.flow_depth == 16):
#			self.flow_frame['storage']['x'] = cv.CreateImage(frame_size,cv.IPL_DEPTH_16S,1)
#			self.flow_frame['storage']['y'] = cv.CreateImage(frame_size,cv.IPL_DEPTH_16S,1)
#			self.flow_scaling = 2**self.flow_depth/(2*self.FLOW_RANGE)
#		elif (self.flow_depth == 32):
#			self.flow_frame['storage']['x'] = self.flow_frame['analysis']['x']
#			self.flow_frame['storage']['y'] = self.flow_frame['analysis']['y']
#			self.flow_scaling = 1.0
#		else:
#			raise Exception("Invalid flow depth value. Valid values are (16, 32).")



#		return
#
#		self.flow_frame['storage']['x'] = self.flow_frame['analysis']['x']
#		self.flow_frame['storage']['y'] = self.flow_frame['analysis']['y']


# ----------------------------------------------------------------------------------------------- #

#	def update_analysis_status(self,status):
#
#		# This method handles the signal 'analysis_canceled' emitted by the class
#		# FlowGui when the flow analysis is canceled through the progress bar.
#
#		self.flow_analysis_canceled = status

# ----------------------------------------------------------------------------------------------- #

	def on_canceled_processing(self):

		# This method handles the signal 'analysis_canceled' emitted by the class
		# FlowGui when the flow analysis is canceled through the progress bar.

		# self.flow_analysis_canceled = status
		self.flow_analysis_canceled = True

# ----------------------------------------------------------------------------------------------- #

#	def write_frame(self):
#
#
#		# TODO: WE MAY WANT TO MOVE THE CODE BELOW TO THE 'write_frame' METHODS OF THE
#		# MovieReaderOpenCV and MovieReaderPyffmpeg CLASSES. THOSE METHODS WOULD THEN
#		# BE RESPONSIBLE FOR CONVERTING THE FLOW FRAMES TO 16 BITS IF NECESSARY BEFORE
#		# WRITING THEM TO DISK.
#
#
#		# If the flow depth is 16, we convert the flow frames from 32 bit float
#		# to 16 bit signed integer and write them to the output file. Otherwise,
#		# we just write the 32 bit float frames to the output file.
#		if (self.flow_depth == 16):
#			cv.ConvertScale(self.flow_frame['analysis']['x'],self.flow_frame['storage']['x'],self.flow_scaling)
#			cv.ConvertScale(self.flow_frame['analysis']['y'],self.flow_frame['storage']['y'],self.flow_scaling)
#		elif (self.flow_depth == 32):
#			cv.Copy(self.flow_frame['analysis']['x'],self.flow_frame['storage']['x'])
#			cv.Copy(self.flow_frame['analysis']['y'],self.flow_frame['storage']['y'])
#		else:
#			raise Exception("Invalid flow depth value. Valid values are (16, 32).")
#
#		# Write flow frames to file
#		self.output_file.write_frame(self.flow_frame['storage'])

# ----------------------------------------------------------------------------------------------- #

	def compute_flow_frame(self):

		# Call the method responsible for computing the flow frame
		self.flow_method['function']()

		# Update the frame number.
		# Important: flow between video frames N and N+1 -> flow frame N
		# self.frame_number += 1
		self.frame_number = self.input_video.frame_number - 1

		# Emit the signal 'flow_frame_computed'
		# self.flow_frame_computed(self.flow_frame,self.frame_number)
		self.flow_frame_processed(self.flow_frame,self.frame_number)

# ----------------------------------------------------------------------------------------------- #

	def optical_flow_hs(self):

		# Use previous (input) velocity field. I dont't know exactly what this means
		use_previous = 0

		# The Lagrangian multiplier.
		# Large values of lambda mean that steep gradients give lower velocities than gentle
		# slopes. In practice, it's a noise filter, low values give you a very sensitive but
		# noisy result, high values filter out the noise, but miss small objects.
		Lambda = 1

		# The termination criteria
		termination_criteria = (cv.CV_TERMCRIT_ITER | cv.CV_TERMCRIT_EPS, 20, 0.3)

		# Perform the optical flow (exactly the same prototype as the C function)
		# cv.CalcOpticalFlowHS(self.video_frame['previous'],self.video_frame['current'],use_previous,self.flow_frame['analysis']['x'],self.flow_frame['analysis']['y'],Lambda,termination_criteria)
		cv.CalcOpticalFlowHS(self.video_frame['previous'],self.video_frame['current'],use_previous,self.flow_frame['x'],self.flow_frame['y'],Lambda,termination_criteria)

# ----------------------------------------------------------------------------------------------- #

	def optical_flow_bm(self):

		# Block size
		block_size = (5, 5)

		# Shift size
		shift_size = (5, 5)

		# Maximum range
		max_range = (5, 5)

		# Use previous (input) velocity field
		use_previous = 0;

		# Perform the optical flow
		# cv.CalcOpticalFlowBM(self.video_frame['previous'],self.video_frame['current'],block_size,shift_size,max_range,use_previous,self.flow_frame['analysis']['x'],self.flow_frame['analysis']['y'])
		cv.CalcOpticalFlowBM(self.video_frame['previous'],self.video_frame['current'],block_size,shift_size,max_range,use_previous,self.flow_frame['x'],self.flow_frame['y'])

# ----------------------------------------------------------------------------------------------- #

	def optical_flow_lk(self):

		# The window size
		window = (5, 5)

		# Perform the optical flow
		# cv.CalcOpticalFlowLK(self.video_frame['previous'],self.video_frame['current'],window,self.flow_frame['analysis']['x'],self.flow_frame['analysis']['y'])
		cv.CalcOpticalFlowLK(self.video_frame['previous'],self.video_frame['current'],window,self.flow_frame['x'],self.flow_frame['y'])

# ----------------------------------------------------------------------------------------------- #

	def pre_process_frame(self):

		# Pre-process frames before they are submitted to optical flow.
		# The original frame is passed as 'frame1', and the processed
		# frame is returned as 'frame2'.

		# Size of the Gaussian kernel used in filtering (in pixels)
		kernel_size = 5

		# Filters the frame
		cv.Smooth(self.video_frame['current'],self.video_frame['current'],cv.CV_GAUSSIAN,kernel_size,0,0)


# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

class FlowFileBinary(object):


# ----------------------------------------------------------------------------------------------- #

	@staticmethod
	def get_dtype_string(flow_depth):

	# IMPORTANT: IN THE MATLAB CODE, THIS FUNCTION IS PRIVATE

	# Information about using strings to construct Numpy dtype objects can be found at:
	# http://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html#arrays-dtypes-constructing
	# Specifically, we used what in the link above is called "Array protocol type strings",
	# where the first character specifies the kind of data and the remaining characters
	# specify how many *bytes* of data.

		if (flow_depth == 16):
			dtype_string = "i2"
		elif (flow_depth == 32):
			dtype_string = "f4"
		else:
			dtype_string = -1

		# Return the dtype string
		return dtype_string

# ----------------------------------------------------------------------------------------------- #


	# This is used to convert the optical flow frames from float to integer.
	# Basically, we map the range [-FLOW_RANGE:+FLOW_RANGE] onto the range
	# [-32768:32767]. Ideally, this value should be computed from the movie
	# under analysis, but this would require a two-pass approach: in the first
	# pass we go through all frames in order to compute the flow range; in the
	# second pass we map each flow frame onto the new range and save it to disk.
	# For now, we use a statically defined flow range. We use the range [-655:655],
	# which yields a quantization step of 0.02 for the flow values, which means
	# the quantization error will be in the range [-0.01:0.01]. This value was
	# chosen after we ran the optical flow analysis on a wide range of movies
	# and found no optical flow value above 600 (absolute value).
	# FLOW_RANGE = 655.0
	flow_range = 655.0

	# The flow depth to be used when saving the optical flow frames to disk
	flow_depth = 16


	# TODO: WE MAY NEED TO SPLIT THIS CLASS INTO TWO SEPARATE CLASSES, NAMELY
	# FLOWFILEREADER AND FLOWFILEWRITER. THE CONSTRUCTOR OF THE FLOWFILEWRITER
	# CLASS WILL TAKE A FLOWANALYSIS OBJECT AS AN INPUT, WHEREAS THE CONSTRUCTOR
	# OF THE FLOWFILEREADER CLASS CAN TAKE JUST A FILE NAME AS INPUT.

	# TODO: WE MAY WANT TO ADD THE FLOWANALYSIS OBJECT AS A PROPERTY OF THE
	# FLOWFILE CLASS

	# TODO: WE NEED TO HIDE SOME METHODS (BY ADDING __ IN FRONT OF THEIR NAMES)

	# Version of the binary flow file (version 0.13 only for the HDF5 format)
	# VERSION = 0.13
	VERSION = 0.12


	# def __init__(self,flow_analysis,file_name,file_mode):
	# def __init__(self,flow_analysis,file_name,file_mode,*args):
	# def __init__(self,file_name,file_mode,*args):
	# def __init__(self,file_name,file_mode,flow_depth,*args):
	# def __init__(self,flow_analysis,file_name,file_mode,flow_depth,*args):
	def __init__(self,file_name,file_mode,flow_depth,*args):

		# Non-hidden attributes
		self.file_name     = None				# Name of the optical flow file
		self.file_header   = None				# Header contents of the optical flow file
		self.fid           = None				# File identifier
		self.file_mode     = None				# File mode (r, w, a, etc)
		self.n_frames      = None				# Number of optical flow frames in the file
		self.flow_frame    = None				# Contents of the current optical flow frame
		self.frame_number  = -1				# Number of the current optical flow frame

		# The name of the optical flow file
		self.file_name = file_name

		# The mode of the optical flow file
		self.file_mode = file_mode

		# Valid flow depths
		valid_flow_depths = (16, 32)

		# If a valid flow depth has been provided, we use it to set the attribute
		# 'self.flow_depth'. Otherwise, the inherited class attribute is used
		if (flow_depth in valid_flow_depths):
			self.flow_depth = flow_depth

		# The format string to be used when reading the binary streams
		self.dtype_string = None

# ----------------------------------------------------------------------------------------------- #

	def open_file(self):

		# Open the optical flow file

		# The valid file modes (only reading and writing for now)
		valid_modes = ("r", "w")

		# Check if the file mode is valid
		if self.file_mode not in valid_modes:
			error_message = "Invalid mode. Valid modes are 'r' and 'w' only."
			raise Exception(error_message)

		# If the file identifier 'fid' is not empty, the file
		# is already open and then we have nothing to do
		if self.fid is not None:
			print "File %s is already open." % self.file_name
			return

		# Try to open the file
		try:
			self.fid = open(self.file_name,self.file_mode)
		except IOError:
			print "Could not open file %s" % self.file_name
			sys.exit(-1)

# ----------------------------------------------------------------------------------------------- #

	def close_file(self):

		# Close the optical flow file

		# If the file identifier 'fid' is an empty matrix, the
		# file is already closed and then we have nothing to do
		if self.fid is None:
			print "File %s is already closed." % self.file_name
			return

		# Try to close the optical flow file. If it succeeds,
		# we assign an empty matrix to the file identifier
		# to signal that the file is closed
		try:
			self.fid.close()
			self.fid = None
		except IOError:
			print "Could not close file %s" % self.file_name
			sys.exit(-1)

# ----------------------------------------------------------------------------------------------- #

	# Important: Apparently, we should avoid using destructors
	# in Python and use contexts ("with" statements) instead

	def __del__(self):

		# This is the destructor function

		# Close the optical flow file
		self.close_file()

# ----------------------------------------------------------------------------------------------- #

	def convert_flow_frame(self,flow_frame):

		# If the flow depth is 16, we convert the flow frames from 32 bit float
		# to 16 bit signed integer and write them to the output file. Otherwise,
		# we just write the 32 bit float frames to the output file.
		if (self.flow_depth == 16):
			# cv.ConvertScale(flow_frame['analysis']['x'],self.flow_frame['x'],self.flow_scaling)
			# cv.ConvertScale(flow_frame['analysis']['y'],self.flow_frame['y'],self.flow_scaling)
			cv.ConvertScale(flow_frame['x'],self.flow_frame['x'],self.flow_scaling)
			cv.ConvertScale(flow_frame['y'],self.flow_frame['y'],self.flow_scaling)
		elif (self.flow_depth == 32):
			# cv.Copy(flow_frame['analysis']['x'],self.flow_frame['x'])
			# cv.Copy(flow_frame['analysis']['y'],self.flow_frame['y'])
			cv.Copy(flow_frame['x'],self.flow_frame['x'])
			cv.Copy(flow_frame['y'],self.flow_frame['y'])
		else:
			raise Exception("Invalid flow depth value. Valid values are (16, 32).")

# ----------------------------------------------------------------------------------------------- #

	def read_frame(self,frame_number):

		# Go to the beginning of the frame of interest
		self.go_to_frame(frame_number)

		# The frame dimensions
		dims = (self.file_header['n_rows'], self.file_header['n_cols'])

		# The number of bytes for each optical flow frame
		n_bytes = self.file_header['image_size']

		# Create an empty dictionary for the flow
		flow = {}

		# Read the x and y flow frames from the binary file
		flow['x'] = numpy.fromstring(self.fid.read(n_bytes),self.dtype_string).reshape(dims).astype("float64")
		flow['y'] = numpy.fromstring(self.fid.read(n_bytes),self.dtype_string).reshape(dims).astype("float64")

		# Scale the flow values back to their original range
		flow['x'] = flow['x']/self.file_header['flow_scaling']
		flow['y'] = flow['y']/self.file_header['flow_scaling']

		# Update the FlowReader object
		self.frame_number = frame_number
		self.flow_frame = flow

# ----------------------------------------------------------------------------------------------- #

	def write_frame(self,flow_frame,frame_number):

		# The input argument 'flow_frame' must be a dictionary with fields 'x' and 'y'

		# Convert the analysis flow frame to the format to be saved to disk
		self.convert_flow_frame(flow_frame)

		# Write the x and y flow frames to the output file
		self.fid.write(self.flow_frame['x'].tostring())
		self.fid.write(self.flow_frame['y'].tostring())

		# Update the frame number
		self.frame_number += 1

# ----------------------------------------------------------------------------------------------- #

	def read_next_frame(self):

		# The next frame to be read
		if self.frame_number is None:
			next_frame = self.file_header['initial_frame']
		else:
			next_frame = self.frame_number+1

		# Read the next frame in the sequence. If no frame has been read so far,
		# we read the first frame of the sequence
		self.read_frame(next_frame)

# ----------------------------------------------------------------------------------------------- #

	#IMPORTANT: IN THE MATLAB CODE, THE FUNCTION BELOW IS PRIVATE

	# Important: 'frame_number' is the number of the *flow* frame in the
	# optical flow file, not the number of the video frame in the original
	# movie. Thus, 'frame_number' = 0 means the first flow frame in flow
	# file, 'frame_number' = 1 means the second flow frame in flow file,
	# and so on.

	def go_to_frame(self,frame_number):

		# Check if the file is open
		# if isempty(reader.fid), disp('File is not open.'); return; end;
		if (self.fid is None):
			print "File is not open."
			return

		# The optical flow limits (in frame numbers)
		# flow_limits = [reader.file_header.initial_frame reader.file_header.final_frame];
		flow_limits = (self.file_header['initial_frame'], self.file_header['final_frame'])

		# These conditions must be met by the variable 'frame_number'
		conditions = [None, None, None]
		conditions[0] = (round(frame_number) == frame_number)
		conditions[1] = (frame_number >= flow_limits[0])
		conditions[2] = (frame_number <= flow_limits[1])

		# The frame number must be a positive integer between 1 and the number of frames in the file
		if not all(conditions):
			error_message = "The frame number must be an integer in the range [%g %g]." % (flow_limits[0], flow_limits[1])
			raise Exception(error_message)

		# Position the file pointer at the beginning of the frame of interest
		# position = reader.header_size+(frame_number-flow_limits(1))*2*reader.frame_size;
		# status = fseek(reader.fid,position,'bof');
		position = self.header_size+(frame_number-flow_limits[0])*2*self.frame_size
		self.fid.seek(position,0)

# ----------------------------------------------------------------------------------------------- #

	def read_file_header(self):

		# Read the header of the binary file. Once the file has been opened for
		# reading, this is the sequence used to read the header:
		#
		# - read the first double (8 bytes), which contains the header version.
		#
		# - read the ninth byte as a character. If it is a valid character, it is
		#   taken to be the file byte order; otherwise, the byte order passed to
		#   the constructor of the class is used instead
		#
		# - now that we know the binary file's byte order, we go back to the beginning
		#   of the file, call the function file_header_layout() passing it the correct
		#   header version, and then we read the whole file header.


		# Go to the beginning of the file
		self.fid.seek(0)

		# Read the first double (8 bytes) from the binary file. Regardless of the header version,
		# the first double in the file will always be the header version itself. Important: the
		# function struct.unpack() always returns a tuple, even if there is only one element.
		# Thus, in order to get the only element from the tuple, we need to index it as [0]
		header_version = struct.unpack('d',self.fid.read(struct.calcsize('d')))[0]

		# Read the ninth byte of the binary file. Starting at version 0.12, this byte contains
		# a character representing the byte order used to write the file to disk. Thus, if this
		# byte contains a valid character (one of 'L', 'l', '<','B', 'b', '>'), this character
		# is taken to be the file byte order and then 'self.byte_order' is updated accordingly.
		# If the byte does not contain a valid character (this will happen for file versions
		# before 0.12), 'self.byte_order' is not modified and, consequently, the byte order
		# passed to the constructor of the class will be used to read the binary file.
		byte_order = struct.unpack('c',self.fid.read(struct.calcsize('c')))[0]
		override_byte_order = False
		if byte_order in ('L', 'l', '<'):
			byte_order = '<'
			override_byte_order = True
		if byte_order in ('B', 'b', '>'):
			byte_order = '>'
			override_byte_order = True

		# If the byte order used to save the file is present in the flow file,
		# it is used to override the byte order passed to the class constructor
		if override_byte_order is True:
			self.byte_order = byte_order
			# print "Flow file contains information about byte order. If byte order information has been passed to the constructor, it will be overriden by the one in the file."
			# warnings.warn("Flow file contains information about byte order. If byte order information has been passed to the constructor, it will be overriden by the one in the file.")

		# Go back to the beginning of the file and re-read the header version,
		# now using the correct byte order (the byte order in the first attempt
		# may have been wrong)
		self.fid.seek(0)
		formatting_string = byte_order + 'd'
		header_version = struct.unpack(formatting_string,self.fid.read(struct.calcsize('d')))[0]

		# The header layout for this version of the file
		# header_items = file_header_layout(header_version)
		header_items = self.file_header_layout(header_version)

		# Create an empty dictionary for the header
		header = {}

		# Go to the beginning of the binary file once again and read the entire header
		self.fid.seek(0)

		# Loop over the header items, reading them from the binary file
		for item in header_items:

			# The name and type of the current header item
			(item_name, item_type) = item

			# The formatting string to be used by struct.unpack()
			# formatting_string = byte_order_char + item_type
			formatting_string = byte_order + item_type

			if (item_name != "first_frame"):

				# If the current item is not the first frame, read its contents from the
				# binary file using the method struct.unpack()
				header[item_name] = struct.unpack(formatting_string,self.fid.read(struct.calcsize(item_type)))[0]

			else:

				# Read the first frame. The only reason we can do this within the for loop
				# is because "first_frame" is the last item in the list "header_items" and,
				# therefore, by the time the else condition below is met, we already know
				# everything we need in order to read the frame contents (more specifically,
				# the "n_rows" and "n_cols" entries of the file header). Note: because we
				# hardcode the image mode ("L") when calling Image.fromstring(), the code
				# below only works for images with one byte/pixel
				n_pixels = header['n_cols']*header['n_rows']
				image_size = (header['n_cols'], header['n_rows'])
				frame_string = self.fid.read(n_pixels)
				# header['first_frame'] = Image.fromstring("L",image_size,frame_string)
				header['first_frame'] = Image.frombytes("L",image_size,frame_string)

		# Store the file header in the FlowFile object
		self.file_header = header

		# Initialize the header and frame sizes, in bytes
		self.header_size = self.fid.tell()
		self.frame_size  = self.file_header['image_size']

		# The format string to be used when reading the binary streams
		self.define_dtype_string()

		# Pre-pend the byte order character to the dtype string so that the binary
		# flow data is read with the correct endiannes from file
		# self.dtype_string = byte_order_char + self.dtype_string
		self.dtype_string = byte_order + self.dtype_string

		# The number of optical flow frames in the file
		self.n_frames = self.file_header['final_frame']-self.file_header['initial_frame']

# ----------------------------------------------------------------------------------------------- #

	def define_dtype_string(self):

	# IMPORTANT: IN THE MATLAB CODE, THIS FUNCTION IS PRIVATE

	# Information about using strings to construct Numpy dtype objects can be found at:
	# http://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html#arrays-dtypes-constructing
	# Specifically, we used what in the link above is called "Array protocol type strings",
	# where the first character specifies the kind of data and the remaining characters
	# specify how many *bytes* of data.

		# Use the bit depth in the file header to define the dtype string to be used
		# when reading the binary flow data
		if (self.file_header['flow_depth'] == 16):
			self.dtype_string = "i2"
		elif (self.file_header['flow_depth'] == 32):
			self.dtype_string = "f4"
		else:
			error_message = "I can only handle 16 and 32 bit optical flow files."
			raise Exception(error_message)

# ----------------------------------------------------------------------------------------------- #

#	def remove_rods(self,rods):
#
#		# TODO: THIS METHOD DOESN'T BELONG IN THIS CLASS. WE HAVE TO MOVE IT SOMEWHERE ELSE
#
#		# Check if there is flow data available
#		if self.flow_frame is None:
#			print "No flow data available. Have any frames been read yet?"
#			return
#
#		# Loop over regions
#		for region in rods:
#
#			# The position of the current region
#			position = region['position']
#
#			# Loop over the flow directions (x and y)
#			for key in self.flow_frame.keys():
#				self.flow_frame[key][position[1]:position[3],position[0]:position[2]] = 0.0

# ----------------------------------------------------------------------------------------------- #

#	def process_rois(self,rois):
#
#		# TODO: THIS METHOD DOESN'T BELONG IN THIS CLASS. WE HAVE TO MOVE IT SOMEWHERE ELSE
#
#		# Check if there is flow data available
#		if self.flow_frame is None:
#			print "No flow data available. Have any frames been read yet?"
#			return
#
#		flow = self.flow_frame
#
#		# Add magnitude channels to the dictionary 'self.flow_frame'
#		flow['x_mag'] = numpy.absolute(flow['x'])
#		flow['y_mag'] = numpy.absolute(flow['y'])
#		flow['mag']   = numpy.sqrt(numpy.square(flow['x']) + numpy.square(flow['y']))
#
#		sums = {}
#
#		# Loop over regions
#		for region in rois:
#
#			# The position of the current region
#			position = region['position']
#
#			# The number of pixels in the current region of interest (for normalization purposes)
#			# n_pixels = (position[3]-position[1]+1)*(position[2]-position[0]+1)
#			n_pixels = (position[3]-position[1])*(position[2]-position[0])
#
#			sums[region['name']] = {}
#
#			# Loop over the flow directions (x and y)
#			for key in flow.keys():
#				# sums[region['name']][key] = numpy.sum(flow[key][position[1]:position[3],position[0]:position[2]])
#				sums[region['name']][key] = numpy.sum(flow[key][position[1]:position[3],position[0]:position[2]])/n_pixels
#
#		# Return the sums
#		return sums

# ----------------------------------------------------------------------------------------------- #

	def init_all(self,flow_analysis):

		# Initialize the flow frames
		self.init_flow_frame(flow_analysis)

		# Initializes the optical flow file. This does two things: i) creates the file
		# header; ii) writes the file header to disk. The second step is very important
		# because it positions the file pointer at the beginning of the file's data area.
		# Note, however, that at this point some fields of the file header will contain
		# garbage, as not all information we need is available. After finishing writing
		# all flow frames to the file, we need to update the file header contents and
		# re-write it to disk.
		# self.output_file.init_flow_file(self)
		# self.output_file.init_flow_file(flow_analysis)
		self.init_flow_file(flow_analysis)

# ----------------------------------------------------------------------------------------------- #

	def init_flow_frame(self,flow_analysis):

		# Initializes the 'flow_frame' attribute
		self.flow_frame = {'x': None, 'y': None}

		# The frame size, in pixels
		# frame_size = (flow_analysis.input_video.width, flow_analysis.input_video.height)
		frame_size = (flow_analysis.width, flow_analysis.height)

		# Create the variables 'flow_x_file' and 'flow_y_file' containing the optical
		# flow values to be saved to disk, stored as either 16 or 32 bit integers. The
		# variables 'flow_x' and 'flow_y' contain the optical flow values to be used
		# in all OpenCV computations, stored as 32-bit floats. If FLOW_DEPTH == 16,
		# 'flow_x_file' and 'flow_y_file' are created as new OpenCV images with a
		# 16-bit depth. If FLOW_DEPTH == 32, 'flow_x_file' and 'flow_y_file' are
		# created as pointers to the variables 'flow_x' and 'flow_y', respectively.
		# Also, define the scale factor by which we have to multiply the float flow
		# frame to map it onto an integer flow frame. This value is passed to the
		# OpenCV function cv.cvConvertScale()
		if (self.flow_depth == 16):
			self.flow_frame['x'] = cv.CreateImage(frame_size,cv.IPL_DEPTH_16S,1)
			self.flow_frame['y'] = cv.CreateImage(frame_size,cv.IPL_DEPTH_16S,1)
			self.flow_scaling = 2**self.flow_depth/(2*self.flow_range)
		elif (self.flow_depth == 32):
			# self.flow_frame['x'] = flow_analysis.flow_frame['analysis']['x']
			# self.flow_frame['y'] = flow_analysis.flow_frame['analysis']['y']
			self.flow_frame['x'] = flow_analysis.flow_frame['x']
			self.flow_frame['y'] = flow_analysis.flow_frame['y']
			self.flow_scaling = 1.0
		else:
			raise Exception("Invalid flow depth value. Valid values are (16, 32).")

# ----------------------------------------------------------------------------------------------- #

	def init_flow_file(self,flow_analysis):

		# Open the optical flow file
		# self.open_file(file_name,mode,file_type)
		self.open_file()

		# Create the header of the flow file
		self.create_header(flow_analysis)

		# Write the file header to disk This is important as it positions
		# the file pointer to the beginning of the data portion of the file
		self.write_file_header()

# ----------------------------------------------------------------------------------------------- #

	def create_header(self,flow_analysis):

		# Create an empty file header. This is a dictionary, where each item is associated
		# with a two element list. The first element of the list is the item value, and the
		# second element is a character representing the precision to be used when writing
		# the element to the output binary file. The characters used to represent precision
		# are those defined in the module 'struct'. This is version 0.12

		# A video frame and a flow frame. We need these in order to
		# extract some attributes
		video_frame = flow_analysis.video_frame['current']
		# flow_frame  = flow_analysis.flow_frame['storage']['x']
		flow_frame  = self.flow_frame['x']

		# Get the file header layout according to the header version
		header_items = self.file_header_layout(self.VERSION)

		# Create an empty dictionary for the header
		header = {}

		# Use the header items to populate the dictionary
		for item in header_items:
			# header[item[0]] = [0, item[1]]
			header[item[0]] = 0

	# Write the machine's byte order to the header file. The binary
	# optical flow file is saved to disk using the machine's native
	# byte order. This information is then used when reading the file
	# to ensure that the correct byte order is preserved. This makes
	# the binary file architecture independent (at least as far as
	# byte order is concerned), i.e., it makes it possible to read
	# the file on a machine with a different byte order than that
	# of the machine where the file was created.
		if (sys.byteorder.lower() == "little"):
			header['byte_order'] = "L"
		else:
			header['byte_order'] = "B"

		# Fill out the header
		header['version']       = self.VERSION										# The version of the flow file
		header['n_rows']        = video_frame.height								# Number of rows in the flow frame
		header['n_cols']        = video_frame.width								# Number of columns in the flow frame
		header['n_channels']    = flow_frame.nChannels							# Number of channels in the flow frame
		# header['flow_depth']  = flow_analysis.flow_depth						# Optical flow depth *in the binary file*
		header['flow_depth']    = self.flow_depth									# Optical flow depth *in the binary file*
		header['frame_rate']    = flow_analysis.input_video.frame_rate		# Video frame rate
		header['initial_frame'] = flow_analysis.initial_frame					# Initial frame
		header['final_frame']   = flow_analysis.final_frame					# Final frame
		header['image_size']    = len(flow_frame.tostring())					# Flow frame size, in bytes
		# header['flow_scaling']  = flow_analysis.flow_scaling				# Value used for scaling the optical flow
		header['flow_scaling']  = self.flow_scaling								# Value used for scaling the optical flow
		header['first_frame']   = video_frame.tostring() + 'z'				# First frame of the video sequence

		# Store the file header in the attribute 'file_header'
		self.file_header = header

# ----------------------------------------------------------------------------------------------- #

	def update_file_header(self,video_frame):

		# Update the output file header with the initial frame read from the input
		# video. This is important because at this point field 'first_frame' of the
		# file header contains garbage, as this information wasn't available when
		# the header was created
		self.file_header['first_frame'] = video_frame.tostring() + 'z'

# ----------------------------------------------------------------------------------------------- #

	def write_file_header(self):

		# Go to the beginning of the file
		self.fid.seek(0)

		# Get the header's item list in the correct order
		# header_items = self.file_header_layout()
		header_items = self.file_header_layout(self.VERSION)

		# Remove the 'first_frame' entry from the header items list. We do this
		# before the loop below because the 'first_frame' item cannot be written
		# to disk using the struct.pack() approach and, therefore, has to be dealt
		# with separately
		header_items.remove(['first_frame', 'B'])

		# Write the header items to the output flow file
		for item in header_items:
			(item_name, item_type) = item
			self.fid.write(struct.pack(item_type,self.file_header[item_name]))

		# Write the 'first_frame' header item to disk
		self.fid.write(self.file_header['first_frame'])

# ----------------------------------------------------------------------------------------------- #

	def file_header_layout(self,version):

#		# The file version
#		version = self.VERSION

		# Important: the variable 'header_items' is defined as a list instead of
		# a dictionary because the order in which the items are laid out in the
		# file is important. As we know, in Python, the keys of a dictionary are
		# not organized in any specific order.

		# Define the header layout according to the header version
		if (version == 0.10):

			header_items = [
			['version',       'd'],					# double
			['n_rows',        'I'],					# unsigned integer
			['n_cols',        'I'],					# unsigned integer
			['n_channels',    'I'],					# unsigned integer
			['flow_depth',    'I'],					# unsigned integer
			['frame_rate',    'd'],					# double
			['initial_frame', 'I'],					# unsigned integer
			['final_frame',   'I'],					# unsigned integer
			['image_size',    'I'],					# unsigned integer
			['n_sigmas',      'I'],					# unsigned integer
			['sigma_x',       'd'],					# double
			['sigma_y',       'd'],					# double
			['sigma_mag',     'd'],					# double
			['first_frame',   'B']]					# unsigned char

		elif (version == 0.11):

			# Changes since 0.10: the addition of the variable 'flow_scaling'

			header_items = [
			['version',       'd'],					# double
			['n_rows',        'I'],					# unsigned integer
			['n_cols',        'I'],					# unsigned integer
			['n_channels',    'I'],					# unsigned integer
			['flow_depth',    'I'],					# unsigned integer
			['frame_rate',    'd'],					# double
			['initial_frame', 'I'],					# unsigned integer
			['final_frame',   'I'],					# unsigned integer
			['image_size',    'I'],					# unsigned integer
			['n_sigmas',      'I'],					# unsigned integer
			['sigma_x',       'd'],					# double
			['sigma_y',       'd'],					# double
			['sigma_mag',     'd'],					# double
			['flow_scaling',  'd'],					# double
			['first_frame',   'B']]					# unsigned char

		elif (version == 0.12):

			# Changes since 0.11: the addition of the variable 'byte_order'

			header_items = [
			['version',       'd'],					# double
			['byte_order',    'c'],					# char (added in version 0.12)
			['n_rows',        'I'],					# unsigned integer
			['n_cols',        'I'],					# unsigned integer
			['n_channels',    'I'],					# unsigned integer
			['flow_depth',    'I'],					# unsigned integer
			['frame_rate',    'd'],					# double
			['initial_frame', 'I'],					# unsigned integer
			['final_frame',   'I'],					# unsigned integer
			['image_size',    'I'],					# unsigned integer
			['n_sigmas',      'I'],					# unsigned integer
			['sigma_x',       'd'],					# double
			['sigma_y',       'd'],					# double
			['sigma_mag',     'd'],					# double
			['flow_scaling',  'd'],					# double
			['first_frame',   'B']]					# unsigned char

		elif (version == 0.13):

			# Changes since 0.12: now the flow file is stored in the HDF5 format,
			# which gives us a lot of flexibility. One of the consequences is that
			# we don't have to worry with the order of the items in the file header
			# anymore, and therefore we can specify the header items as a dictionary
			# suitable to be used by the create_table() constructor in the PyTables
			# package. Furthermore, since we can now add and/or remove header items
			# without having to worry about the file header layout on disk, we wont
			# need to increment the file header version anymore

			header_items = {
			'version'				: tables.Float64Col(),					# signed 64-bit float
			'input_video_file'	: tables.StringCol(itemsize=1),		# 1 character string (will be changed later)
			'n_rows'					: tables.UInt32Col(),					# unsigned 32-bit integer
			'n_cols'					: tables.UInt32Col(),					# unsigned 32-bit integer
			'n_channels'			: tables.UInt32Col(),					# unsigned 32-bit integer
			'flow_depth'			: tables.UInt32Col(),					# unsigned 32-bit integer
			'frame_rate'			: tables.Float64Col(),					# signed 64-bit float
			'initial_frame'		: tables.UInt32Col(),					# unsigned 32-bit integer
			'final_frame'			: tables.UInt32Col(),					# unsigned 32-bit integer
			'image_size'			: tables.UInt32Col(),					# unsigned 32-bit integer
#			'n_sigmas'				: tables.UInt32Col(),					# unsigned 32-bit integer
			'sigma_x'				: tables.Float64Col(),					# signed 64-bit float
			'sigma_y'				: tables.Float64Col(),					# signed 64-bit float
			'sigma_mag'				: tables.Float64Col(),					# signed 64-bit float
			'flow_scaling'		: tables.Float64Col(),					# signed 64-bit float
			'first_frame'			: tables.StringCol(itemsize=1)}		# 1 character string (will be changed later)

		else:

			# error(sprintf('I don''t know how to read file header version %g',version));
			# print "I don't know how to read file header version %g" % version
			# sys.exit(-1)
			error_message = "I don't know how to read file header version %g" % version
			raise Exception(error_message)

		# Return the header items
		return header_items

# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

class FlowFileHDF(FlowFileBinary):


	# TODO: I THINK THIS CLASS IS PRETTY MUCH FINISHED. WE HAVE TO CHECK IT. ONE THING
	# THAT STILL NEEDS TO BE DONE IS TO UPDATE THE FRAME READING METHODS.


	# Version of the binary flow file (version 0.13 only for the HDF5 format)
	VERSION = 0.13

	# Some class variables controlling the format (string or numpy) and the
	# compression level of the output HDF file
	OUTPUT_FILE_TYPE  = "string"               # either "string" or "numpy"
	# OUTPUT_FILE_TYPE  = "numpy"               # either "string" or "numpy"
	# COMPRESSION_LIB   = "zlib"                 # either "zlib" or "bzip2"
	# COMPRESSION_LEVEL = 2


# ----------------------------------------------------------------------------------------------- #

	def __init__(self,file_name,file_mode,flow_depth,compression_method="zlib",compression_level=2):

		# We don't want to simply override the superclass' constructor, but rather
		# extend it. To do so, we need to explicitly call the superclass' method
		# and only then add the extra code. Otherwise, we end up with an object
		# improperly initialized
		# super(FlowFileHDF,self).__init__(flow_analysis,file_name,file_mode)
		# super(FlowFileHDF,self).__init__(file_name,file_mode)
		# super(FlowFileHDF,self).__init__(flow_analysis,file_name,file_mode,flow_depth)
		super(FlowFileHDF,self).__init__(file_name,file_mode,flow_depth)

		# Add a new attribute to the class with some compression parameters
		self.compression = {'method': compression_method, 'level': compression_level}

# ----------------------------------------------------------------------------------------------- #

	def open_file(self):

		# Open the optical flow file

		# The valid file modes (only reading and writing for now)
		valid_modes = ("r", "w")

		# Check if the file mode is valid
		if self.file_mode not in valid_modes:
			error_message = "Invalid mode. Valid modes are 'r' and 'w' only."
			raise Exception(error_message)

		# If the file identifier 'fid' is not empty, the file
		# is already open and then we have nothing to do
		if self.fid is not None:
			print "File %s is already open." % self.file_name
			return

		# Try to open the file
		try:
			self.fid = tables.open_file(self.file_name,self.file_mode)
		except IOError:
			print "Could not open file %s" % file_name
			sys.exit(-1)

# ----------------------------------------------------------------------------------------------- #

	# Important: Apparently, we should avoid using destructors
	# in Python and use contexts ("with" statements) instead

	def __del__(self):

		# This is the destructor function

		# Close the optical flow file
		self.close_file()

# ----------------------------------------------------------------------------------------------- #

	def read_frame(self,frame_number):

		# The frame dimensions
		dims = (self.file_header['n_rows'], self.file_header['n_cols'])

		# The number of bytes for each optical flow frame
		n_bytes = self.file_header['image_size']

		# Create an empty dictionary for the flow
		flow = {}

		# The index used to access the correct frame in the data array
		array_index = frame_number - self.file_header['initial_frame']

		# Read the flow values from the CArray. Interestingly, this works not only for strings,
		# but for Numpy arrays as well. Weird or what?
		# flow['x'] = numpy.fromstring(self.fid.root.flow.x[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float64")
		# flow['y'] = numpy.fromstring(self.fid.root.flow.y[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float64")
		# flow['x'] = numpy.fromstring(self.fid.root.flow.x[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float32")
		# flow['y'] = numpy.fromstring(self.fid.root.flow.y[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float32")
		flow['x'] = numpy.fromstring(self.fid.root.flow.x[array_index][0:n_bytes],self.dtype_string).reshape(dims).astype("float32")
		flow['y'] = numpy.fromstring(self.fid.root.flow.y[array_index][0:n_bytes],self.dtype_string).reshape(dims).astype("float32")

		# Scale the flow values back to their original range
		flow['x'] = flow['x']/self.file_header['flow_scaling']
		flow['y'] = flow['y']/self.file_header['flow_scaling']

		# Update the FlowFile object
		self.frame_number = frame_number
		self.flow_frame = flow

# ----------------------------------------------------------------------------------------------- #

	def read_raw_frame(self,frame_number):

		# This function is similar to the function read_frame() above. The difference
		# is that this function returns the raw optical flow frame data, whereas the
		# function read_frame() returns the frame data nicely organized into a numpy
		# array with the correct dimensions and data type. This function was written
		# in order to be used with the Python FlowServer class which communicates with
		# a Java FlowClient object over a TCP connection using the ZMQ library. The
		# FlowServer object opens a flow field HDF file, reads the optical flow data
		# one frame at a time and sends it to the FlowClient object. Initially, the
		# FlowServer object was using the function read_frame() to read the frames,
		# but since the server had to convert the numpy arrays returned by read_frame()
		# to strings in order to send them to the client over the TCP connection, this
		# involved a lot of unnecessary computations. By returning the raw frame data,
		# this function prevents this problem. Now, the FlowServer object can read the
		# raw frame data and pass it directly to the client over the TCP connection.
		# Important: since this function simply reads the raw data from the HDF flow
		# file and returns it to the caller, the normalization required for recovering
		# the correct optical flow vectors is not done here. Thus, this will have to
		# be done by the FlowClient object.

		# The frame dimensions
		dims = (self.file_header['n_rows'], self.file_header['n_cols'])

		# The number of bytes for each optical flow frame
		n_bytes = self.file_header['image_size']

		# Create an empty dictionary for the flow
		flow = {}

		# The index used to access the correct frame in the data array
		array_index = frame_number - self.file_header['initial_frame']

		# Read the flow values from the CArray. Interestingly, this works not only for strings,
		# but for Numpy arrays as well. Weird or what?
		# flow['x'] = numpy.fromstring(self.fid.root.flow.x[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float64")
		# flow['y'] = numpy.fromstring(self.fid.root.flow.y[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float64")
		# flow['x'] = numpy.fromstring(self.fid.root.flow.x[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float32")
		# flow['y'] = numpy.fromstring(self.fid.root.flow.y[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float32")
		# flow['x'] = numpy.fromstring(self.fid.root.flow.x[array_index][0:n_bytes],self.dtype_string).reshape(dims).astype("float32")
		# flow['y'] = numpy.fromstring(self.fid.root.flow.y[array_index][0:n_bytes],self.dtype_string).reshape(dims).astype("float32")
		flow['x'] = self.fid.root.flow.x[array_index][0:n_bytes]
		flow['y'] = self.fid.root.flow.y[array_index][0:n_bytes]

		# # Scale the flow values back to their original range
		# flow['x'] = flow['x']/self.file_header['flow_scaling']
		# flow['y'] = flow['y']/self.file_header['flow_scaling']

		# # Update the FlowFile object
		# self.frame_number = frame_number
		# self.flow_frame = flow

		# Return the raw frame data
		return flow

# ----------------------------------------------------------------------------------------------- #

	def write_frame(self,flow_frame,frame_number):

		# The input argument 'flow_frame' must be a dictionary with fields 'x' and 'y'

		# Convert the analysis flow frame to the format to be saved to disk
		self.convert_flow_frame(flow_frame)

		# The row number in the CArray table where the flow frame will be written.
		# Please note that we ignore the frame number passed by the FlowAnalysis
		# object and instead we get it from the 'frame_number' attribute. The two
		# values are the same anyway.
		frame_number = self.frame_number + 1

		# Write the flow frames to disk according to the output file type
		# (either "string" or "numpy")
		if (self.OUTPUT_FILE_TYPE == "string"):

			# Very important: when assigning the contents of the optical flow frames
			# (the byte stream returned by the tostring() method of the IplImage object),
			# for some reason trailing '\x00' elements in the string being assigned are
			# NOT copied over to the CArray in the output HDF file. I had a long discussion
			# with the PyTables developer Francesc Alted about this issue and he said that,
			# although this doesn't seem to be expected behaviour, he didn't recognize this
			# as a bug, but rather suggested me to stop using strings to store binary data
			# (this was kind of surprising to me actually, as I thought strings were *the*
			# way of storing binary data in Python 2.x). At least for the time being, we
			# work around this by adding an extra character at the end of the string in
			# order to ensure that the last character will not be an '\x00' and, therefore,
			# the string will be copied over to the CArray intact. Obviously, when reading
			# these strings from the HDF file, we have to drop the last character before
			# converting the string back to either an IplImage or a Numpy array.
			self.fid.root.flow.x[frame_number] = self.flow_frame['x'].tostring() + 'z'
			self.fid.root.flow.y[frame_number] = self.flow_frame['y'].tostring() + 'z'

		elif (self.OUTPUT_FILE_TYPE == "numpy"):

			self.fid.root.flow.x[frame_number,:,:] = numpy.asarray(cv.GetMat(self.flow_frame['x']))
			self.fid.root.flow.y[frame_number,:,:] = numpy.asarray(cv.GetMat(self.flow_frame['y']))

		else:

			raise Exception("Invalid value for self.OUTPUT_FILE_TYPE.")

		# Update the frame counter
		self.frame_number = frame_number

# ----------------------------------------------------------------------------------------------- #

	def read_file_header(self):

		# The table containing the "file header". This table contains only one row
		header_table = self.fid.root.header.header_table

		# The number of rows in the header table. There should be only one, but if
		# there are more, we read the last one, as it should be the most up to date
		n_rows = self.fid.root.header.header_table.nrows

		# The number of the row to be read
		row_number = n_rows - 1

		# Initialize the property 'file_header' as an empty dictionary
		self.file_header = {}

		# Loop over the colums in the table 'header_table', reading the
		# header items and populating the dictionary self.file_header
		for header_item in header_table.colnames:
			# self.file_header[header_item] = header_table[0][header_item]
			self.file_header[header_item] = header_table[row_number][header_item]

		# The format string to be used when reading the binary streams
		self.define_dtype_string()

		# The number of optical flow frames in the file
		self.n_frames = self.file_header['final_frame']-self.file_header['initial_frame']

		# Convert the first frame from string to a PIL image
		image_size = (self.file_header['n_cols'], self.file_header['n_rows'])
		# self.file_header['first_frame'] = Image.fromstring("L",image_size,self.file_header['first_frame'])
		# self.file_header['first_frame'] = Image.frombytes("L",image_size,self.file_header['first_frame'])
		self.file_header['first_frame'] = Image.frombytes("L",image_size,self.file_header['first_frame'][0:-1])

# ----------------------------------------------------------------------------------------------- #

	# TODO: THIS METHOD HAS BEEN COPIED FROM THE CLASS FlowFileBinary().
	# DO WE NEED TO PASS A FlowAnalysis OBJECT TO THIS METHOD? I THINK THE
	# ONLY REASON FOR DOING THAT RIGHT NOW IS BECAUSE WE NEED THE FIRST
	# FRAME OF THE VIDEO IN ORDER TO STORE IT IN THE HEADER. I THINK WE
	# CAN LEAVE ALL OTHER FIELDS BLANK, AS WE RE-WRITE THE HEADER AT THE
	# END OF THE ANALYSIS ANYWAY IN ORDER TO UPDATE SOME FIELDS, AS THE
	# STANDARD DEVIATIONS.


	def create_header(self,flow_analysis):

		# Create an empty file header. This is a dictionary, where each item is associated
		# with a two element list. The first element of the list is the item value, and the
		# second element is a character representing the precision to be used when writing
		# the element to the output binary file. The characters used to represent precision
		# are those defined in the module 'struct'. This is version 0.12


		# The name of the input video file
		input_video_file = flow_analysis.input_video.input_file

		# A video frame and a flow frame. We need these in order to
		# extract some attributes
		video_frame = flow_analysis.video_frame['current']
		flow_frame  = self.flow_frame['x']

		# Get the file header layout according to the header version
		# header_items = self.file_header_layout()
		header_items = self.file_header_layout(self.VERSION)

		# The number of bytes in a video frame. We need to know this in order
		# to allocate the correct space in the header table
		# n_bytes = len(flow_analysis.video_frame['current'].tostring())
		# n_bytes = len(video_frame.tostring())
		n_bytes = len(video_frame.tostring()) + 1

		# Update the key 'first_frame' in the dictionary 'header_items'
		# with the correct number of bytes
		header_items['first_frame'] = tables.StringCol(itemsize=n_bytes)

		# Number of characters in the name of the input video file. We need to
		# know this in order to allocate the correct space in the header table
		n_chars = len(input_video_file)

		# Update the key 'input_video_file' in the dictionary 'header_items'
		# with the correct number of characters
		header_items['input_video_file'] = tables.StringCol(itemsize=n_chars)

		# Create an empty dictionary for the header
		header = {}

		# Use the header items to populate the dictionary
		for item in header_items:
			header[item] = None

		# Fill out the header
		header['version']          = self.VERSION										# The version of the flow file
		header['input_video_file'] = input_video_file								# Name of the input video file
		header['n_rows']           = video_frame.height								# Number of rows in the flow frame
		header['n_cols']           = video_frame.width								# Number of columns in the flow frame
		header['n_channels']       = flow_frame.nChannels							# Number of channels in the flow frame
		# header['flow_depth']     = flow_analysis.flow_depth						# Optical flow depth *in the binary file*
		header['flow_depth']       = self.flow_depth									# Optical flow depth *in the binary file*
		header['frame_rate']       = flow_analysis.input_video.frame_rate		# Video frame rate
		header['initial_frame']    = flow_analysis.initial_frame					# Initial frame
		header['final_frame']      = flow_analysis.final_frame					# Final frame
		header['image_size']       = len(flow_frame.tostring())					# Flow frame size, in bytes
		# header['flow_scaling']   = flow_analysis.flow_scaling					# Value used for scaling the optical flow
		header['flow_scaling']     = self.flow_scaling								# Value used for scaling the optical flow
		header['first_frame']      = video_frame.tostring() + 'z'				# First frame of the video sequence

		# Store the file header in the attribute 'file_header'
		self.file_header = header

		# Return the variable 'header_items'. We use this to create a group in the
		# HDF file for storing the file header
		return header_items

# ----------------------------------------------------------------------------------------------- #

	def write_file_header(self):

		# Loop over the header items, writing them to the corresponding items
		# in the header table in the HDF file
		for item_name in self.fid.root.header.header_table.colnames:
			self.fid.root.header.header_table.row[item_name] = self.file_header[item_name]

		# Insert the row (the only one we use) into the table and flush it to disk
		self.fid.root.header.header_table.row.append()
		self.fid.root.header.header_table.flush()

# ----------------------------------------------------------------------------------------------- #

	def init_flow_file(self,flow_analysis):

		# Open the optical flow file
		# self.open_file(file_name,mode,file_type)
		self.open_file()

		# Create HDF groups for the file header and the flow vectors
		self.fid.create_group("/", 'header', 'File Header')
		self.fid.create_group("/", 'flow', 'Optical Flow Vectors')

		# Create the file header and write it to disk
		header_items = self.create_header(flow_analysis)

		# Create a table under the group 'header' for storing the file header
		self.fid.create_table(self.fid.root.header,'header_table',header_items,"File Header")

#		# Write the file header to disk
#		self.write_file_header()

		# An optical flow frame. We extract frame size and the data size
		# information from this
		# flow_frame = flow_analysis.flow_frame['storage']['x']
		flow_frame = self.flow_frame['x']

		# The number of flow frames to be stored in the output flow file
		n_flow_frames = flow_analysis.final_frame - flow_analysis.initial_frame

		# Determines the CArray's shape and atom according to the output file
		# type (either "string" or "numpy")
		if (self.OUTPUT_FILE_TYPE == "string"):

			# The shape and atom of the CArrays that will hold the optical flow frames
			string_length = len(flow_frame.tostring())+1
			array_shape = (n_flow_frames,)
			array_atom = tables.StringAtom(itemsize=string_length)

		elif (self.OUTPUT_FILE_TYPE == "numpy"):

			# The shape and atom of the CArrays that will hold the optical flow frames.
			# Very important: the shape of the CArray has a HUGE impact on the speed the
			# HDF file can be written to disk, especially if compression is being used.
			# Using (height,width,n_frames) is *much* slower than (n_frames,height,width).
			# In a test where 100 flow frames (720 x 480 pixels) were saved to disk, the
			# difference was 5.0223 times when no compression was used, and 35.3863 times
			# when compression was used. Apparently, this is caused by the way the data
			# is laid out on disk. According to the Pytables developer (Francesc Alted),
			# data on disk is written in C-order and therefore we must ensure that the
			# leading dimensions in the CArray vary the slowest.
			# array_shape = (flow_frame.height,flow_frame.width,n_flow_frames)
			array_shape = (n_flow_frames,flow_frame.height,flow_frame.width)
			aux = numpy.asarray(cv.GetMat(flow_frame))
			array_atom = tables.Atom.from_dtype(aux.dtype)

		else:

			raise Exception("Invalid value for self.OUTPUT_FILE_TYPE.")

		# Define a filter object in order to set some compression parameters for the CArrays
		# array_filters = tables.Filters(complib=self.COMPRESSION_LIB,complevel=self.COMPRESSION_LEVEL)
		array_filters = tables.Filters(complib=self.compression['method'],complevel=self.compression['level'])

		# Create CArrays to store the horizontal and vertical optical flow frames
		self.fid.create_carray(self.fid.root.flow,'x',array_atom,array_shape,filters=array_filters)
		self.fid.create_carray(self.fid.root.flow,'y',array_atom,array_shape,filters=array_filters)

# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

class FlowApplication(object):

	# The output file types
	output_file_types = ('flow_field', 'flow_regions', 'flow_signals')

	# Define a dictionary with the flow file types. This will be used to
	# populate the flow file type combobox. The first element in the list
	# below will be the default file type
#	output_file_types = [
#		{'name': "binary", 'label': 'Binary', 'function': FlowFileBinary},
#		{'name': "hdf",    'label': 'HDF',    'function': FlowFileHDF}]
	flow_file_types = [
		{'name': "binary", 'label': 'Binary', 'function': FlowFileBinary, 'file_extension': ".bin"},
		{'name': "hdf",    'label': 'HDF',    'function': FlowFileHDF,    'file_extension': ".hdf"}]

# ----------------------------------------------------------------------------------------------- #

	#----- Signals emitted by this class -----#

	# Signal emitted when the optical flow analysis is canceled
	analysis_canceled = notify.Signal()

	# Signal emitted after an optical flow frame has been computed
	# flow_frame_computed = notify.Signal()
	flow_frame_processed = notify.Signal()

# ----------------------------------------------------------------------------------------------- #

	@staticmethod
	def get_default_output_file_type():

		return "hdf"

# ----------------------------------------------------------------------------------------------- #

	def __init__(self, input_file_name, flow_regions = None, output_flow_file_name = None,
		initial_frame = None, final_frame = None, flow_method = None, flow_depth = 16,
		video_backend = "opencv", output_flow_file_type = "hdf"):

		self.input_file_name          = input_file_name
		self.flow_source              = None
		# self.flow_regions             = flow_regions
		self.flow_signals             = None
		# self.output_signals_file_name = output_signals_file_name
		self.output_flow_file_name    = output_flow_file_name
		self.output_flow_file_type    = None
		self.output_flow_file         = None
		self.initial_frame            = initial_frame
		self.final_frame              = final_frame
		self.flow_method              = flow_method
		self.flow_depth               = flow_depth
		self.video_backend            = video_backend

		# Initialize the output file type to its default value
		self.set_output_flow_file_type(output_flow_file_type)

		# Set the optical flow source (either a FlowAnalysis or a FlowReader object)
		self.set_flow_source()

		# Set the frame range
		self.set_frame_range()

		# Set the flow signals
		self.set_flow_signals(flow_regions)

		# Set the output flow file
		self.set_output_flow_file(self.flow_source, self.output_flow_file_name, self.flow_depth, self.output_flow_file_type["name"])


#		# Set the output flow signals file
#		self.set_output_signals_file()


# ----------------------------------------------------------------------------------------------- #

	def set_flow_source(self):

		# The extension of the input file
		(zzz,extension) = os.path.splitext(self.input_file_name)

		# Flow file extensions
		flow_file_extensions = [file_type['file_extension'] for file_type in FlowApplication.flow_file_types]

		# Define the flow source (either a FlowAnalysis or a FlowReader object)
		# according to the extension of the input file
		if extension.lower() in flow_file_extensions:
			self.flow_source = FlowReader(self.input_file_name)
			# If the flow source is an optical flow file, we must set the attribute
			# 'output_flow_file_name' to None in order to signal that no optical flow
			# file should be created
			self.output_flow_file_name = None
		else:
			self.flow_source = FlowAnalysis(self.input_file_name, self.flow_method, self.video_backend)


		# ----- Signal/slot connections ----- #

		# This connection allows the flow analysis status to be updated
		# when the analysis is canceled through the progress bar
		self.analysis_canceled.connect(self.flow_source.on_canceled_processing)

		# This connection allows the progress bar to be updated every time
		# a flow frame is computed
		self.flow_source.flow_frame_processed.connect(self.flow_frame_processed)

# ----------------------------------------------------------------------------------------------- #

	def set_frame_range(self):

		# Set the frame range to be used in the analysis. If we don't do this,
		# the whole video is analyzed
		self.flow_source.set_frame_range(self.initial_frame, self.final_frame)

		# Update the 'initial_frame' and 'final_frame' attributes according
		# to the values in the 'flow_source' object. This is important when
		# using None values in the FlowApplication object to signal that
		# default values should be used for the initial and final frames.
		# In this case, we have to get the actual numerical values of the
		# default frame numbers from the 'flow_source' object.
		self.initial_frame = self.flow_source.initial_frame
		self.final_frame   = self.flow_source.final_frame

# ----------------------------------------------------------------------------------------------- #

	def set_output_flow_file_type(self, type_name = None):

		# If no file type has been provided, we use the default one
		if (type_name == None):
			type_name = self.get_default_output_file_type()

		try:
			# The file type dictionary that matches the provided type name 'type_name'
			(file_type,) = [file_type for file_type in self.flow_file_types if file_type['name'] == type_name]
		except:
			error_message = "Invalid output file type: %s." % type_name
			raise Exception(error_message)

		# Update the file type dictionary in the FlowAnalysis object
		# self.output_file_type = file_type
		self.output_flow_file_type = file_type

# ----------------------------------------------------------------------------------------------- #

#	def __get_signals_file_name(self, flow_file_name):
#
#		# If no flow file name has been provided, we return a default signals file name
#		if not flow_file_name:
#			# (flow_file_name, regions_file_name, signals_file_name) = self.__get_default_output_file_names()
#			output_file_names = self.__get_default_output_file_names()
#			# return signals_file_name
#			return output_file_names
#
#		# Split the full path to the flow file into a directory name and a relative file name
#		(dir_name, file_name) = os.path.split(flow_file_name)
#
#		# If the flow file name contains the string "_flow_field" (this is the case
#		# for default flow file names), we remove it
#		file_name = file_name.replace("_flow_field","")
#
#		# Construct the name of the signals file by adding the string "_flow_signals" to
#		# the variable 'file_name'
#		(file_prefix, file_suffix) = os.path.splitext(file_name)
#		signals_file_name = file_prefix + "_flow_signals" + file_suffix
#
#		# Turn the path to the signals file into an absolute path
#		signals_file_name = os.path.join(dir_name, signals_file_name)
#
#		# Return the name of the signals file
#		return signals_file_name

# ----------------------------------------------------------------------------------------------- #

#	def __get_regions_file_name(self, flow_file_name):
#
#		# If no flow file name has been provided, we return a default signals file name
#		if not flow_file_name:
#			(flow_file_name, regions_file_name, signals_file_name) = self.__get_default_output_file_names()
#			return regions_file_name
#
#		# Split the full path to the flow file into a directory name and a relative file name
#		(dir_name, file_name) = os.path.split(flow_file_name)
#
#		# If the flow file name contains the string "_flow_field" (this is the case
#		# for default flow file names), we remove it
#		file_name = file_name.replace("_flow_field","")
#
#		# Construct the name of the signals file by adding the string "_flow_signals" to
#		# the variable 'file_name'
#		(file_prefix, file_suffix) = os.path.splitext(file_name)
#		signals_file_name = file_prefix + "_flow_signals" + file_suffix
#
#		# Turn the path to the signals file into an absolute path
#		signals_file_name = os.path.join(dir_name, signals_file_name)
#
#		# Return the name of the signals file
#		return signals_file_name

# ----------------------------------------------------------------------------------------------- #

#	def __get_default_output_flow_file_name(self):
#
#		# Initialize the output file name to an empty matrix, which will be
#		# returned to the calling function if no input file has been defined
#		# so far
#		output_file_name = ""
#
#		input_file_name = self.input_file_name
#		initial_frame   = self.initial_frame
#		final_frame     = self.final_frame
#
#		# Searches for the last occurrence of the
#		# character '.' in the file name string
#		dot_position = input_file_name.rfind(".")
#
#		# The file name suffix (file extension)
#		# suffix = ".bin"
#		suffix = self.output_flow_file_type['file_extension']
#
#		# The file name prefix
#		if (dot_position == -1):
#			prefix = input_file_name
#		else:
#			prefix = input_file_name[:dot_position]
#
#		# Append the frame number to the prefix
#		# prefix = prefix + "_frames_" + str(initial_frame) + "_" + str(final_frame)
#		prefix = prefix + "_frames_" + str(initial_frame) + "_" + str(final_frame) + "_flow_field"
#
#		# The output file name
#		output_file_name = prefix + suffix
#
#		# Returns the output file name
#		return output_file_name

# ----------------------------------------------------------------------------------------------- #

#	def __get_default_output_file_names(self):
#
#		input_file_name = self.input_file_name
#		initial_frame   = self.initial_frame
#		final_frame     = self.final_frame
#
#		# The file types
#		# file_types = ('flow_field', 'flow_regions', 'flow_signals')
#		file_types = self.output_file_types
#
#		# The file name suffixes
#		file_suffixes = {}
#		file_suffixes['flow_field'] = self.output_flow_file_type['file_extension']
#		file_suffixes['flow_regions'] = ".xls"
#		file_suffixes['flow_signals'] = ".hdf"
#
#		# The prefix of the input file name
#		(prefix, zzz) = os.path.splitext(input_file_name)
#
#		# Append frame numbers to the prefix
#		prefix = prefix + "_frames_" + str(initial_frame) + "_" + str(final_frame)
#
#		# Initialize a dictionary with the file names
#		file_names = {}
#
#		# The file name for each file type
#		for file_type in file_types:
#			file_names[file_type] = prefix + "_" + file_type + file_suffixes[file_type]
#
#		# Return the output file names
#		return file_names

# ----------------------------------------------------------------------------------------------- #

	def __get_default_flow_file_name(self):

		input_file_name = self.input_file_name
		initial_frame   = self.initial_frame
		final_frame     = self.final_frame

		# The suffix to be used for the output flow file
		suffix = self.output_flow_file_type['file_extension']

		# The prefix of the input file name
		(prefix, zzz) = os.path.splitext(input_file_name)

		# Append frame numbers to the prefix
		prefix = prefix + "_frames_" + str(initial_frame) + "_" + str(final_frame)

		# The complete name of the output flow file
		flow_file_name = prefix + '_flow_field' + suffix

		# Return the name of the flow file
		return flow_file_name

# ----------------------------------------------------------------------------------------------- #

	def __get_output_file_names(self, flow_file_name):

		# If no flow file name has been provided, we assign it a default value
		if not flow_file_name:
			flow_file_name = self.__get_default_flow_file_name()

		# The file name suffixes
		file_suffixes = {}
		file_suffixes['flow_regions'] = ".xls"
		file_suffixes['flow_signals'] = ".hdf"

		# Split the full path to the flow file into a directory name and a relative file name
		(dir_name, file_name) = os.path.split(flow_file_name)

		# If the flow file name contains the string "_flow_field" (this is the case
		# for default flow file names), we remove it
		file_name = file_name.replace('_flow_field','')

		# Split the flow file name in its prefix and suffix (extension) parts
		(prefix, suffix) = os.path.splitext(file_name)

		# Initialize a dictionary with the output file names
		output_file_names = {}

		# The file name for each file type
		for file_type in file_suffixes.keys():
			output_file_names[file_type] = prefix + "_" + file_type + file_suffixes[file_type]
			output_file_names[file_type] = os.path.join(dir_name, output_file_names[file_type])

		# Add the name of the flow file to the 'output_file_names' dictionary
		# output_file_names['flow_field'] = flow_file_name

		# Return the output file names
		return output_file_names

# ----------------------------------------------------------------------------------------------- #

	def set_output_flow_file(self, flow_analysis, output_flow_file_name = None, flow_depth = 16,
		output_flow_file_type = "hdf", compression_method = "zlib", compression_level = 2):

		# The variable 'output_flow_file_name' controls whether or not an output file
		# will be defined and, if so, what its name will be. There are 3 possibilities:
		# i) if its value is None, no output file will be defined; ii) if its value is
		# the empty string "", an output file with a default file name will be used;
		# iii) if its value is any other string, an output file will be defined whose
		# name is the string's contents.

		# Define the output file type. We need to do that before defining the
		# output file name, as the latter (more specifically, its extension)
		# will depend on the former in the case of a default file name
		# self.set_output_file_type(output_file_type)
		self.set_output_flow_file_type(output_flow_file_type)

		# If 'output_flow_file_name' is an empty string, we assign it the default
		# output flow file name
		if (output_flow_file_name == ""):
			# output_flow_file_name = self.__get_default_output_file_name()
			# output_flow_file_name = self.__get_default_output_flow_file_name()
			output_flow_file_name = self.__get_default_flow_file_name()

		# Update the output file name in the FlowApplication object
		# self.output_file_name = output_file_name
		self.output_flow_file_name = output_flow_file_name

		# If 'output_flow_file_name' is None, there's nothing else to do
		if (output_flow_file_name == None):
			return

		# Create flow file writer and attach it to the attribute "output_file". The class of the
		# output file is determined by the field 'function' of the output file type dictionary
		# output_file = self.output_file_type['function'](output_file_name,"w",flow_depth,compression_method,compression_level)
		output_flow_file = self.output_flow_file_type['function'](output_flow_file_name,"w",flow_depth,compression_method,compression_level)

		# Initializes the optical flow file. This does two things: i) creates the file
		# header; ii) writes the file header to disk. The second step is very important
		# because it positions the file pointer at the beginning of the file's data area.
		# Note, however, that at this point some fields of the file header will contain
		# garbage, as not all information we need is available. After finishing writing
		# all flow frames to the file, we need to update the file header contents and
		# re-write it to disk.
		output_flow_file.init_all(flow_analysis)


		# ----- Signal/slot connections ----- #

		# This connection allows the output file's header to be updated after
		# the initial frame of the analysis is read. This is necessary because
		# the initial frame is stored in the header of the output file
		# flow_analysis.initial_frame_read.connect(output_flow_file.update_file_header)
		flow_analysis.initial_video_frame_read.connect(output_flow_file.update_file_header)

		# This connection will cause each computed flow frame to be written
		# to the output flow file
		# flow_analysis.flow_frame_computed.connect(output_flow_file.write_frame)
		flow_analysis.flow_frame_processed.connect(output_flow_file.write_frame)

		# Define the methods in the output flow file to be called when the
		# analysis is finished. The order of the connections below is very
		# important. The methods will be executed in the same order as the
		# connections are defined.
		# flow_analysis.flow_analysis_finished.connect(output_flow_file.write_file_header)
		# flow_analysis.flow_analysis_finished.connect(output_flow_file.close_file)
		flow_analysis.flow_source_finished.connect(output_flow_file.write_file_header)
		flow_analysis.flow_source_finished.connect(output_flow_file.close_file)


		# ----------------------------------- #


		# Update the 'output_file' attribute
		self.output_flow_file = output_flow_file

# ----------------------------------------------------------------------------------------------- #

	def set_output_signals_file(self):


		# The attribute 'flow_regions' can take 3 different values: i) if None, no regions
		# of interest (and therefore no flow signals) are defined; ii) if [] (an empty list),
		# a default region of interest encompassing the whole image frame will be defined;
		# iii) if a FlowRegions object, the regions of interest defined in the object will
		# be used

		if (self.flow_regions == None):
			return

		if (self.flow_regions == []):

			# The frame size
			frame_size = (flow_source.input_video.width, flow_source.input_video.height)

			# Create a FlowRegions object and add a default region of interest to it
			self.flow_regions = FlowRegions(frame_size)
			self.flow_regions.add_default_roi()





		# The flow source
		flow_source = self.flow_source


		# Default output file names
		(flow_file_name, signals_file_name) = self.__get_default_output_file_names()






		# Create the FlowSignalsFile object
		output_signals_file = FlowSignalsFile(signals_file_name, "w")

		# output_signals_file.init_file(self.flow_regions)
		output_signals_file.init_file(flow_source, self.flow_regions)




#		# Create the FlowSignals object
#		# self.flow_signals = FlowSignals(self.flow_regions, self.initial_frame, self.final_frame, flow_source.input_video.frame_rate)
#		self.flow_signals = FlowSignals(self.flow_regions, flow_source.initial_frame, flow_source.final_frame, flow_source.input_video.frame_rate)






		# ----- Signal/slot connections ----- #

		# This connection will cause each computed flow frame to have its regions
		# of interested processed
		# self.flow_source.flow_frame_processed.connect(self.flow_signals.process_flow_frame)
		flow_source.flow_frame_processed.connect(output_signals_file.process_flow_frame)


		# flow_source.flow_source_finished.connect(output_flow_file.write_file_header)
		# flow_source.flow_source_finished.connect(output_flow_file.close_file)
		flow_source.flow_source_finished.connect(output_signals_file.close_file)




		# Update the 'output_signals_file' attribute
		self.output_signals_file = output_signals_file


		return




















		# Create flow file writer and attach it to the attribute "output_file". The class of the
		# output file is determined by the field 'function' of the output file type dictionary
		# output_file = self.output_file_type['function'](output_file_name,"w",flow_depth,compression_method,compression_level)
		output_flow_file = self.output_flow_file_type['function'](output_flow_file_name,"w",flow_depth,compression_method,compression_level)

		# Initializes the optical flow file. This does two things: i) creates the file
		# header; ii) writes the file header to disk. The second step is very important
		# because it positions the file pointer at the beginning of the file's data area.
		# Note, however, that at this point some fields of the file header will contain
		# garbage, as not all information we need is available. After finishing writing
		# all flow frames to the file, we need to update the file header contents and
		# re-write it to disk.
		output_flow_file.init_all(flow_analysis)


		# ----- Signal/slot connections ----- #

		# This connection allows the output file's header to be updated after
		# the initial frame of the analysis is read. This is necessary because
		# the initial frame is stored in the header of the output file
		# flow_analysis.initial_frame_read.connect(output_flow_file.update_file_header)
		flow_analysis.initial_video_frame_read.connect(output_flow_file.update_file_header)

		# This connection will cause each computed flow frame to be written
		# to the output flow file
		# flow_analysis.flow_frame_computed.connect(output_flow_file.write_frame)
		flow_analysis.flow_frame_processed.connect(output_flow_file.write_frame)

		# Define the methods in the output flow file to be called when the
		# analysis is finished. The order of the connections below is very
		# important. The methods will be executed in the same order as the
		# connections are defined.
		# flow_analysis.flow_analysis_finished.connect(output_flow_file.write_file_header)
		# flow_analysis.flow_analysis_finished.connect(output_flow_file.close_file)
		flow_analysis.flow_source_finished.connect(output_flow_file.write_file_header)
		flow_analysis.flow_source_finished.connect(output_flow_file.close_file)


		# ----------------------------------- #


		# Update the 'output_file' attribute
		self.output_flow_file = output_flow_file





# ----------------------------------------------------------------------------------------------- #

	def set_flow_signals(self, flow_regions):

		# The attribute 'flow_regions' can take 3 different values: i) if None, no regions
		# of interest (and therefore no flow signals) are defined; ii) if [] (an empty list),
		# a default region of interest encompassing the whole image frame will be defined;
		# iii) if a FlowRegions object, the regions of interest defined in the object will
		# be used

		# If no regions of interest have been defined (flow_regions == None), there are
		# no flow signals
		if (flow_regions == None):
			return

		# If 'flow_regions' is an empty list, we define a default flow region encompassing
		# the whole image frame
		if (flow_regions == []):

			# The frame size
			# frame_size = (self.flow_source.input_video.width, self.flow_source.input_video.height)
			frame_size = (self.flow_source.width, self.flow_source.height)

			# Create a FlowRegions object and add a default region of interest to it
			flow_regions = FlowRegions(frame_size)
			flow_regions.add_default_roi()


		# Create the FlowSignals object and initialize it for the flow analysis
		# self.flow_signals = FlowSignals(self.flow_regions, self.initial_frame, self.final_frame, flow_source.input_video.frame_rate)
		# self.flow_signals = FlowSignals(self.flow_regions, self.flow_source.initial_frame, self.flow_source.final_frame, self.flow_source.input_video.frame_rate)
		# self.flow_signals = FlowSignals(flow_regions, self.flow_source.initial_frame, self.flow_source.final_frame, self.flow_source.input_video.frame_rate)
		# self.flow_signals = FlowSignals(flow_regions, self.flow_source.initial_frame, self.flow_source.final_frame, self.flow_source.frame_rate)
		self.flow_signals = FlowSignals()
		self.flow_signals.init_for_analysis(flow_regions, self.flow_source.initial_frame, self.flow_source.final_frame, self.flow_source.frame_rate)

		# ----- Signal/slot connections ----- #

		# This connection will cause each computed flow frame to have its regions
		# of interested processed
		self.flow_source.flow_frame_processed.connect(self.flow_signals.process_flow_frame)

# ----------------------------------------------------------------------------------------------- #

	def plot_flow_signals(self, signal_names = ('x', 'y', 'mag'), show_figure = True, formats = ('png', 'pdf')):

		if self.flow_signals:

			# Initially set 'output_file_names' to None, which means that no figure files
			# will be saved to disk. We later check if 'formats' have been provided and, if
			# so, we find the appropriate output file names
			output_file_names = None

			# If 'formats' have been provided, that means that figure files will be saved
			# to disk. In that case, we find the names of the files to be saved, one per
			# file format provided in 'formats'. Note: the figure files have all the same
			# names, just with different extensions
			if formats:

				# Get the name of the signals file from the name of the flow file
				output_file_names = self.__get_output_file_names(self.output_flow_file_name)
				signals_file_name = output_file_names['flow_signals']

				# The file name prefix. File names for the different file types will
				# be constructed by adding the appropriate extension to this prefix
				(file_prefix, extension) = os.path.splitext(signals_file_name)

				# The number of output files
				n_files = len(formats)

				# Initialize the list of output file names
				output_file_names = [None]*n_files

				# Loop over the output files
				for indice in range(n_files):

					# The file name for the current file format
					output_file_names[indice] = file_prefix + '.' + formats[indice]

			# Plot the flow signals and save the figure to the different file formats
			self.flow_signals.plot_signals(signal_names, show_figure, output_file_names)

# ----------------------------------------------------------------------------------------------- #

	def save_flow_signals(self, file_types = None):

		if self.flow_signals:

			# Get the name of the signals file from the name of the flow file
			output_file_names = self.__get_output_file_names(self.output_flow_file_name)
			signals_file_name = output_file_names['flow_signals']

			# Save the flow signals to disk
			self.flow_signals.save_signal_files(signals_file_name, file_types)

# ----------------------------------------------------------------------------------------------- #

	def save_flow_regions(self):

		if self.flow_signals.flow_regions:

			# Get the name of the signals file from the name of the flow file
			output_file_names = self.__get_output_file_names(self.output_flow_file_name)
			regions_file_name = output_file_names['flow_regions']

			# Save the flow regions to disk
			self.flow_signals.flow_regions.save_to_file(regions_file_name)

# ----------------------------------------------------------------------------------------------- #

	def run(self, show_plots = True):

		# Save the flow regions to disk
		self.save_flow_regions()

		# Run the flow analysis
		self.flow_source.run()

		# Plot the signal files
		# self.plot_flow_signals()
		# self.plot_flow_signals(signal_names = ('x', 'y', 'mag'))
		self.plot_flow_signals(signal_names = ('x', 'y', 'mag'), show_figure = show_plots)


		# Save the flow signals to disk
		# self.save_flow_signals()
		# self.save_flow_signals(['hdf'])
		# self.save_flow_signals(['xls'])
		# self.save_flow_signals(['hdf', 'pdf'])
		# self.save_flow_signals(['hdf', 'xls', 'pdf'])
		self.save_flow_signals(['hdf', 'xls'])

# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

class FlowRegions(object):

# ----------------------------------------------------------------------------------------------- #

	def __init__(self, frame_size = (640, 480)):

		# The frame size
		(self.frame_width, self.frame_height) = frame_size

		# The regions of interest and disinterest
		self.rois          = []
		self.rods          = []

# ----------------------------------------------------------------------------------------------- #

	def add_default_roi(self):

		# Name of the default region of interest
		region_name = "frame"

		# Position of the default region of interest
		region_position = (0, 0, self.frame_width, self.frame_height)

		# Add the default region of interest to the FlowRegions object
		self.add_roi(region_name, region_position)

# ----------------------------------------------------------------------------------------------- #

	def add_roi(self, region_name, region_position):

		# The names of the current regions
		region_names = [region["name"] for region in self.rois]

		# Check whether the region being added exists already
		if (region_name in region_names):
			print "Region '%s' cannot be added because it already exists." % region_name
			return -1

		# A dictionary with the current region's properties
		region = {}
		region["name"]     = region_name
		region["position"] = region_position

		# Append the new region to the list of regions in the FlowReader object
		self.rois.append(region)

		# Return the created region
		return region

# ----------------------------------------------------------------------------------------------- #

	def add_rod(self, region_name, region_position):

		# A dictionary with the current region's properties
		region = {}
		region["name"]     = region_name
		region["position"] = region_position

		# Append the new region to the list of regions in the FlowReader object
		self.rods.append(region)

		# Return the created region
		return region

# ----------------------------------------------------------------------------------------------- #

	def add_region(self, region_name, region_type, region_position):

		# Assign the variable 'regions' to the correct attribute (either self.rois
		# or self.rods) according to the region type
		if (region_type == "interest"):
			# regions = self.rois
			self.add_roi(region_name, region_position)
		elif (region_type == "disinterest"):
			# regions = self.rods
			self.add_rod(region_name, region_position)
		else:
			error_message = "Invalid region type. Valid types are 'interest' and 'disinterest'."
			raise Exception(error_message)


# ----------------------------------------------------------------------------------------------- #

	def save_to_file(self, file_name):

		# Create a work book
		work_book = xlwt.Workbook()

		# Add a sheet to the work book
		regions_sheet = work_book.add_sheet('flow regions')

		# The cell titles
		cell_titles = ('Region name', 'Region type', 'x1', 'y1', 'x2', 'y2')

		# Write the cell titles to the Excel file
		for col_number in range(len(cell_titles)):
			regions_sheet.write(0, col_number, cell_titles[col_number])

		# The initial row in the Excel file where to start writing the flow regions
		row_number = 2

		# Loop over the regions of interest
		for region in self.rois:

			# The contents to be written to the current row
			region_position = [int(number) for number in region['position']]
			# row_contents = (region['name'], 'interest') + region['position']
			# row_contents = (region['name'], 'interest') + region_position
			row_contents = [region['name'], 'interest'] + region_position

			# Loop over the cells in the current row, writing each of them to the Excel file
			for col_number in range(len(row_contents)):
					regions_sheet.write(row_number, col_number, row_contents[col_number])

			# Increment the row number
			row_number += 1

		# Loop over the regions of disinterest
		for region in self.rods:

			# The contents to be written to the current row
			row_contents = (region['name'], 'disinterest') + tuple(region['position'])

			# Loop over the cells in the current row, writing each of them to the Excel file
			for col_number in range(len(row_contents)):
				regions_sheet.write(row_number, col_number, row_contents[col_number])

			# Increment the row number
			row_number += 1

		# Save the file to disk
		work_book.save(file_name)

# ----------------------------------------------------------------------------------------------- #

	def load_from_file(self, input_file_name, sheet_number=0):

		# Create a work book
		work_book = xlrd.open_workbook(input_file_name)

		# The sheet with the regions of interest
		regions_sheet = work_book.sheets()[sheet_number]

		# The cell titles
		cell_titles = ('Region name', 'Region type', 'x1', 'y1', 'x2', 'y2')

		# The header row
		header_row = regions_sheet.row_values(0)

		# Check if this is a flow regions file
		if False in [i==j for (i, j) in zip(cell_titles,header_row)]:
			print "This doesn't seem to be a flow regions file. Quitting..."
			return

		# Loop over the rows of the sheet
		for row in range(2,regions_sheet.nrows):

			# The contents of the current row
			row_contents = regions_sheet.row_values(row)

			# Extract the name, type and position of the current region
			region_name     = row_contents[0]
			region_type     = row_contents[1]
			region_position = row_contents[2:6]

			# Add the current region to the FlowRegions object
			self.add_region(region_name, region_type, region_position)

# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

class FlowSignals(object):


	#----- Signals emitted by this class -----#

#	# Signal emitted when the optical flow analysis is canceled
#	analysis_canceled = notify.Signal()
#
#	# Signal emitted after an optical flow frame has been computed
#	flow_frame_computed = notify.Signal()

# ----------------------------------------------------------------------------------------------- #

#	def __init__(self, flow_regions, initial_frame, final_frame, frame_rate):
#
#
#
#		# TODO: IF WE WANT TO ADD A METHOD TO THIS CLASS FOR READING THE MOTION
#		# SIGNALS FROM HDF/XLS MOTION FILES, WE'LL PROBABLY NEED TWO CONSTRUCTORS,
#		# SINCE THE CLASS WILL HAVE TO BE INITIALIZED DIFFERENTLY FOR EACH CASE.
#		# APPARENTLY, THE CORRECT WAY OF DOING THIS IS BY USING CLASS METHODS.
#		# I READ ABOUT THIS SOMEWHERE ONCE.
#
#
#		self.flow_regions  = flow_regions
#		self.initial_frame = initial_frame
#		self.final_frame   = final_frame
#		self.n_frames      = final_frame - initial_frame
#		# self.n_frames      = final_frame - initial_frame + 1
#		self.frame_rate    = frame_rate
#		self.time_vector   = None
#		self.signals       = {}
#
#		# The names of the regions of interest
#		region_names = [region['name'] for region in flow_regions.rois]
#
#		# The signal names
#		signal_names = ["x", "y", "x_mag", "y_mag", "mag"]
#
#		# The time vector. If we don't explicitly set the type to float here,
#		# we get a sequence of integers and, if the frame rate happens to be
#		# an integer as well, we end up with a sequence of integers for the
#		# time vector, which is wrong.
#		self.time_vector = numpy.arange(self.initial_frame, self.final_frame, dtype=numpy.float)
#		# self.time_vector = numpy.linspace(self.initial_frame, self.final_frame,self.n_frames)
#		self.time_vector = self.time_vector/self.frame_rate
#
#		# Initialize the signals
#		for region_name in region_names:
#			self.signals[region_name] = {}
#			for signal_name in signal_names:
#				# self.signals[region_name][signal_name] = [None]*self.n_frames
#				# self.signals[region_name][signal_name] = numpy.zeros(self.n_frames)
#				aux = numpy.empty(self.n_frames); aux.fill(numpy.nan)
#				self.signals[region_name][signal_name] = aux

# ----------------------------------------------------------------------------------------------- #

	def __init__(self):

		self.flow_regions  = None
		self.time_vector   = None
		self.initial_time  = None
		self.sampling_rate = None
		self.n_samples     = None
		self.signals       = {}

# ----------------------------------------------------------------------------------------------- #

	def init_for_analysis(self, flow_regions, initial_frame, final_frame, frame_rate):

		# Fill some properties
		self.flow_regions  = flow_regions
		self.sampling_rate = frame_rate
		self.initial_time  = initial_frame/float(frame_rate)
		self.n_samples     = final_frame - initial_frame

		# The names of the regions of interest
		region_names = [region['name'] for region in flow_regions.rois]

		# The signal names
		signal_names = ["x", "y", "x_mag", "y_mag", "mag"]

		# The time vector. If we don't explicitly set the type to float here,
		# we get a sequence of integers and, if the frame rate happens to be
		# an integer as well, we end up with a sequence of integers for the
		# time vector, which is wrong.
		# self.time_vector = numpy.linspace(self.initial_frame, self.final_frame,self.n_frames)
		# self.time_vector = numpy.arange(self.initial_frame, self.final_frame, dtype=numpy.float)
		self.time_vector = numpy.arange(initial_frame, final_frame, dtype=numpy.float)
		# self.time_vector = self.time_vector/self.frame_rate
		self.time_vector = self.time_vector/self.sampling_rate

		# Initialize the signals
		for region_name in region_names:
			self.signals[region_name] = {}
			for signal_name in signal_names:
				# self.signals[region_name][signal_name] = [None]*self.n_frames
				# self.signals[region_name][signal_name] = numpy.zeros(self.n_frames)
				# aux = numpy.empty(self.n_frames); aux.fill(numpy.nan)
				aux = numpy.empty(self.n_samples); aux.fill(numpy.nan)
				self.signals[region_name][signal_name] = aux

# ----------------------------------------------------------------------------------------------- #
#
#	def add_roi(self, region_name, region_position):
#
#		# The names of the current regions
#		region_names = [region["name"] for region in self.rois]
#
#		# Check whether the region being added exists already
#		if (region_name in region_names):
#			print "Region '%s' cannot be added because it exists already." % region_name
#			return -1
#
#		# A dictionary with the current region's properties
#		region = {}
#		region["name"]     = region_name
#		region["position"] = region_position
#
#		# Append the new region to the list of regions in the FlowReader object
#		self.rois.append(region)
#
#		# Create a signal with the same name as the region in self.signals
#		signal_names = ["x", "y", "x_mag", "y_mag", "mag"]
#		self.signals[region_name] = {}
#		for signal_name in signal_names:
#			# self.signals[region_name][signal_name] = [None]*self.n_frames
#			self.signals[region_name][signal_name] = numpy.zeros(self.n_frames)
#
#		# Return the created region
#		return region
#
## ----------------------------------------------------------------------------------------------- #
#
#	def add_rod(self, region_name, region_position):
#
#		# A dictionary with the current region's properties
#		region = {}
#		region["name"]     = region_name
#		region["position"] = region_position
#
#		# Append the new region to the list of regions in the FlowReader object
#		self.rods.append(region)
#
#		# Return the created region
#		return region
#
## ----------------------------------------------------------------------------------------------- #
#
#	def add_region(self, region_name, region_type, region_position):
#
#		# Assign the variable 'regions' to the correct attribute (either self.rois
#		# or self.rods) according to the region type
#		if (region_type == "interest"):
#			# regions = self.rois
#			self.add_roi(region_name, region_position)
#		elif (region_type == "disinterest"):
#			# regions = self.rods
#			self.add_rod(region_name, region_position)
#		else:
#			error_message = "Invalid region type. Valid types are 'interest' and 'disinterest'."
#			raise Exception(error_message)
#
#
#
#		return
#
#
#
#
#		# If the region being added is a region of interest, we have to make
#		# sure that no region of interest with the same name already exists
#		if (region_type == "interest"):
#
#			# The names of the current regions
#			region_names = [region["name"] for region in regions]
#
#			# Check whether the region being added exists already
#			if (region_name in region_names):
#				print "Region '%s' cannot be added because it exists already." % region_name
#				return -1
#
#		# A dictionary with the current region's properties
#		region = {}
#		region["name"]     = region_name
#		region["type"]     = region_type
#		region["position"] = region_position
#
#		# Append the new region to the list of regions in the FlowReader object
#		regions.append(region)
#
#		# Create a signal with the same name as the region in self.signals
#		signal_names = ["x", "y", "x_mag", "y_mag", "mag"]
#		self.signals[region_name] = {}
#		for signal_name in signal_names:
#			self.signals[region_name][signal_name] = [None]*self.n_frames
#
#		# Return the created region
#		return region
#
# ----------------------------------------------------------------------------------------------- #

	def process_rods(self,flow_frame):

		# Loop over regions
		# for region in self.rods:
		for region in self.flow_regions.rods:

			# The position of the current region
			position = region['position']

			# Loop over the flow directions (x and y)
			for key in flow_frame.keys():
				flow_frame[key][position[1]:position[3],position[0]:position[2]] = 0.0

# ----------------------------------------------------------------------------------------------- #

	def process_rois(self,flow_frame,frame_number):

		# IMPORTANT: REGION COORDINATES ARE SPECIFIED AS [X1, Y1, X2, Y2]

		# Add magnitude channels to the dictionary 'self.flow_frame'
		flow_frame['x_mag'] = numpy.absolute(flow_frame['x'])
		flow_frame['y_mag'] = numpy.absolute(flow_frame['y'])
		flow_frame['mag']   = numpy.sqrt(numpy.square(flow_frame['x']) + numpy.square(flow_frame['y']))

		# The index used to access the correct frame in the data array
		# array_index = frame_number - self.initial_frame
		initial_frame = int(round(self.initial_time*self.sampling_rate))
		array_index = frame_number - initial_frame

		# Loop over regions
		# for region in self.rois:
		for region in self.flow_regions.rois:

			# The name of the current position
			region_name = region['name']

			# The position of the current region
			position = region['position']

			# The number of pixels in the current region of interest (for normalization purposes)
			# n_pixels = (position[3]-position[1]+1)*(position[2]-position[0]+1)
			n_pixels = (position[3]-position[1])*(position[2]-position[0])

			# Loop over the flow signals
			for signal_name in flow_frame.keys():
				# sums[region['name']][key] = numpy.sum(flow[key][position[1]:position[3],position[0]:position[2]])
				# self.signals[region_name][signal_name][frame_number] = numpy.sum(flow_frame[signal_name][position[1]:position[3],position[0]:position[2]])/n_pixels
				self.signals[region_name][signal_name][array_index] = numpy.sum(flow_frame[signal_name][position[1]:position[3],position[0]:position[2]])/n_pixels

# ----------------------------------------------------------------------------------------------- #

	# def process_flow_frame(self, flow_frame, frame_number):
	def process_flow_frame(self, frame, frame_number):


		# IMPORTANT: We have a problem here. If the flow frame is coming from the
		# FlowAnalysis class, it comes as an OpenCV image (iplimage), whereas if
		# it comes from the FlowReader class, it comes as a numpy array (the
		# 'read_frame' method of both FlowFileBinary and FlowFileHDF classes
		# converts the OpenCV image to a numpy array using numpy.fromstring().
		# As a quick workaround, we check for that here and convert the flow
		# frame to a numpy array in case it is in the OpenCV format. In the
		# long term, I think the best solution is to drop the use of OpenCV
		# images and work only with numpy arrays (however, I'm not totally
		# sure that's possible; I know OpenCV these days tend to favor the
		# use of CVMat images instead of IplImages, but I'm not sure the
		# CVMat is a numpy format).


#		# HERE'S HOW IT IS DONE WHEN READING A BIN FILE
#		# Read the x and y flow frames from the binary file
#		flow['x'] = numpy.fromstring(self.fid.read(n_bytes),self.dtype_string).reshape(dims).astype("float64")
#		flow['y'] = numpy.fromstring(self.fid.read(n_bytes),self.dtype_string).reshape(dims).astype("float64")
#
#		# HERE'S HOW IT IS DONE WHEN READING AN HDF FILE
#		# Read the flow values from the CArray. Interestingly, this works not only for strings,
#		# but for Numpy arrays as well. Weird or what?
#		flow['x'] = numpy.fromstring(self.fid.root.flow.x[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float64")
#		flow['y'] = numpy.fromstring(self.fid.root.flow.y[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float64")

		# cv.iplimage
		# cv.cvmat


		# We need to create a new dictionary variable 'flow_frame' and then copy the
		# contents of 'frame' to the new variable in order to avoid changing the
		# contents of the original variable. If we don't do that, the contents of
		# the original variable 'frame' will change even in the scope of the calling
		# function, and we definetily don't want that, especially because we do a
		# conversion from IplImage to cvMat here
		flow_frame = {}

		# If the flow frames are IplImages, we convert them to Numpy arrays. If the
		# flow frames are coming from a FlowReader object which is reading the flow
		# information from file, the flow frames will be Numpy arrays as both the
		# FlowFileBinary and FlowFileHDF classes do the conversion upon reading the
		# information from disk. On the other hand, if the flow frames are coming
		# from the FlowAnalysis class, they will be IplImages and therefore we have
		# to do the conversion here.
		# TODO: I think the best long term solution is to work with cvMat arrays
		# instead of IplImages throughout our code. However, in order to do that,
		# we need to update the writing and reading of flow frames to/from disk.
		# if (type(flow_frame["x"]) == cv.iplimage):
		if (type(frame["x"]) == cv.iplimage):

			# flow_depth = flow_frame["x"].depth
			flow_depth = frame["x"].depth
			dtype_string = FlowFileBinary.get_dtype_string(flow_depth)

			# The frame dimensions
			# dims = (flow_frame["x"].height, flow_frame["x"].width)
			dims = (frame["x"].height, frame["x"].width)


		# flow_frame_x = numpy.fromstring(flow_frame["x"].tostring(),dtype_string).reshape(dims).astype("float64")
		# flow_frame_y = numpy.fromstring(flow_frame["y"].tostring(),dtype_string).reshape(dims).astype("float64")
		# flow_frame["x"] = numpy.fromstring(flow_frame["x"].tostring(),dtype_string).reshape(dims).astype("float64")
		# flow_frame["y"] = numpy.fromstring(flow_frame["y"].tostring(),dtype_string).reshape(dims).astype("float64")
		flow_frame["x"] = numpy.fromstring(frame["x"].tostring(),dtype_string).reshape(dims).astype("float64")
		flow_frame["y"] = numpy.fromstring(frame["y"].tostring(),dtype_string).reshape(dims).astype("float64")


		# Process the regions of disinterest. Important: these regions must
		# be processed *before* the regions of interest
		# if self.rods:
		if self.flow_regions.rods:
			self.process_rods(flow_frame)

		# Process the regions of disinterest.
		# if self.rois:
		if self.flow_regions.rois:
			self.process_rois(flow_frame,frame_number)

# ----------------------------------------------------------------------------------------------- #

	def save_signal_files(self, file_name, file_types = None):

		# The valid file types
		# valid_file_types = ("hdf", "xls", "pdf")
		valid_file_types = ("hdf", "xls")

		# If 'file_types' is None, it defaults to all valid file types
		if file_types is None:
			file_types = valid_file_types

		# A dictionary with the functions used to write each output file type
		writing_functions = {}
		writing_functions['hdf'] = self.write_to_hdf_file
		writing_functions['xls'] = self.write_to_xls_file
		# writing_functions['pdf'] = self.write_to_pdf_file

		# Remove invalid file types (if any) from the variable 'file_types'
		file_types = list(set(valid_file_types).intersection(file_types))

		# The file name prefix. File names for the different file types will
		# be constructed by adding the appropriate extension to this prefix
		(file_prefix, extension) = os.path.splitext(file_name)

		# Loop over the file types
		for file_type in file_types:

			# The file name for the current file type
			file_name = file_prefix + '.' + file_type

			# Write file of the current type to disk
			writing_functions[file_type](file_name)

			# Print a message confirming the file has been saved to disk
			print "Motion signals saved to the file %s." % file_name

# ----------------------------------------------------------------------------------------------- #

	def plot_signals(self, signal_names = ('x', 'y', 'mag'), show_figure = True, output_file_names = None):

		# Plot the flow signals and optionally saves the plotted figure to disk.
		# The argument 'signal_names' is a list with the names of the signals to be plotted, if
		# it is None, all signals are plotted. The argument 'show_figure' is a boolean controling
		# whether or not the figure should be shown on screen; when used in combination with the
		# argument 'output_file_names', it can be useful for saving figures to disk without having
		# to show them on screen. The argument 'output_file_names' is a list of file names to which
		# to save the figure. If it is None, no figure is saved to disk. So, there are basically
		# three use cases of interest:
		#
		# - show_figure = True; output_file_names = None => figure shown, no files saved
		# - show_figure = True; output_file_names not None => figure shown, files saved
		# - show_figure = False; output_file_names not None => figure not shown, files saved

		# plt.ion()

		# The region and signal names
		# region_names = signals.keys()
		region_names = self.signals.keys()
		region_names.sort()
		# all_signal_names = signals[region_names[0]].keys()
		all_signal_names = self.signals[region_names[0]].keys()
		all_signal_names.sort()

		# If the input parameter 'signal_names' has not been provided,
		# it defaults to all signals
		if not signal_names:
			signal_names = all_signal_names

		# Make sure that 'signal_names' contains only valid signal names
		signal_names = [name for name in signal_names if name in all_signal_names]
		signal_names.sort()

		# The plot legends for all signals
		legendas = {'mag': 'Magnitude','x_mag': 'Horz. Magnitude','y_mag': 'Vert. Magnitude','x': 'Horizontal','y': 'Vertical'}

		# Discard legends for the signals that are not in 'signal_names'
		legendas = dict([(name, legendas[name]) for name in signal_names])

		# The number of regions
		n_regions = len(region_names)

		# The number of signals
		n_signals = len(signal_names)

		# The number of samples in the motion signals
		# n_samples = signals[region_names[0]][signal_names[0]].shape[0]
		n_samples = self.signals[region_names[0]][signal_names[0]].shape[0]

		# A matrix with the signals to be plotted, one signal per column
		signal_matrix = numpy.zeros((n_samples,n_signals))

		# Initialize the axes handles vector
		axes_handles = [None]*n_regions

		# Create the figure where to plot the signals
		figure_handle = plt.figure()

		# Loop over regions, plotting the signals for each region in a separate panel (subplot)
		for m in range(n_regions):

			# Important: at least when using pyplot, the subplot number is one-based
			# (as opposed to Python's zero-based numbering system)
			# axes_handles[m] = plt.subplot(n_regions,1,m+1)
			axes_handles[m] = figure_handle.add_subplot(n_regions,1,m+1)
			# axes_handles[m] = figure_handle.add_subplot(n_regions,1,m+1,sharex=axes_handles[m-1])

			for n in range(n_signals):
				# signal_matrix[:,n] = signals[region_names[m]][signal_names[n]]
				signal_matrix[:,n] = self.signals[region_names[m]][signal_names[n]]

			# Plot all signals for the current region on the current axes
			# plt.plot(time_vector,signal_matrix,'.-')
			# axes_handles[m].plot(time_vector,signal_matrix,'.-')
			axes_handles[m].plot(self.time_vector,signal_matrix,'.-')

			# Set the ylabel of the current axes
			#	ylabel(region_names{m},'Interpreter','none');
			axes_handles[m].set_ylabel(region_names[m])

			# Add a legend to the current axes
			axes_handles[m].legend([legendas[name] for name in signal_names])

			#	if (m<n_regions), set(gca,'XTickLabel',[]); end;
			if (m<n_regions-1):
				axes_handles[m].set_xticklabels("")

		#% Link all axes along the horizontal direction
		#linkaxes(axes_handles,'x');
		#
		# Set the xlabel for the last axes only
		axes_handles[-1].set_xlabel('Time (s)')

		# Show the figure on screen
		if show_figure:
			# plt.show(block=False)
			plt.show()

		# Save the figure to file
		if output_file_names:
			for file_name in output_file_names:
				figure_handle.savefig(file_name)

		# Return the figure handle to the calling function
		return figure_handle

# ----------------------------------------------------------------------------------------------- #

	def write_to_hdf_file(self, file_name):

		# The region and signal names
		# region_names = signals.keys()
		region_names = self.signals.keys()
		region_names.sort()
		# signal_names = signals[region_names[0]].keys()
		signal_names = self.signals[region_names[0]].keys()
		signal_names.sort()

		# Open the output file for writing
		fid = tables.open_file(file_name,"w")

		# Create a group for storing the flow signals
		sums_group = fid.create_group("/", 'sums', 'Regions')

		# The array shape and type in the HDF file
		# array_shape = time_vector.shape
		# array_type = time_vector.dtype
		array_shape = self.time_vector.shape
		array_type = self.time_vector.dtype
		array_atom = tables.Atom.from_dtype(array_type)

		# CArray filters
		compression_method = "zlib"
		#compression_method = "bzip2"
		#compression_level = 0
		compression_level = 2
		array_filters = tables.Filters(complib=compression_method,complevel=compression_level)

		# Create a CArray for storing the time vector
		array = fid.create_carray("/",'time_vector',array_atom,array_shape,filters=array_filters)

		# Write the time vector to a CArray
		# array[:] = time_vector
		array[:] = self.time_vector

		# The plot legends for all signals
		legendas = {'mag': 'Magnitude','x_mag': 'Horz. Magnitude','y_mag': 'Vert. Magnitude','x': 'Horizontal','y': 'Vertical'}

		# The number of regions
		n_regions = len(region_names)

		# The number of signals
		n_signals = len(signal_names)

		# The number of samples in the motion signals
		# n_samples = signals[region_names[0]][signal_names[0]].shape[0]
		# n_samples = len(time_vector)
		n_samples = len(self.time_vector)

		# Initialize a vector where to store the region groups
		region_groups = [None]*n_regions

		# Loop over regions, plotting the signals for each region in a separate panel (subplot)
		for m in range(n_regions):

			# Create a group in the HDF file for the current region
			region_groups[m] = fid.create_group(sums_group,region_names[m],region_names[m])

			# Loop over signals in the current region
			for n in range(n_signals):

				# Create an array in the HDF file for the current signal
				array = fid.create_carray(region_groups[m],signal_names[n],array_atom,array_shape,filters=array_filters)

				# Write the current signal to the current CArray
				# array[:] = signals[region_names[m]][signal_names[n]]
				array[:] = self.signals[region_names[m]][signal_names[n]]

		# Close the file
		fid.close()

# ----------------------------------------------------------------------------------------------- #

#	def read_from_file(self, file_name):
#
#		# IMPORTANT: THIS METHOD READS THE SIGNALS FROM THE PROVIDED HDF FILE.
#		# HOWEVER, SINCE NO INFORMATION ABOUT THE REGIONS OF INTEREST BESIDES
#		# THEIR NAMES IS STORED IN THE HDF FLOW SIGNALS FILE, THE FlowSignals
#		# OBJECT CANNOT BE RE-CONSTRUCTED COMPLETELY. IN THE FUTURE, WE MAY
#		# WANT TO MODIFY THE METHOD FOR SAVING FLOW SIGNALS FILE IN ORDER
#		# TO SAVE THE FLOW REGIONS ALONG WITH THE SIGNALS THEMSELVES IN
#		# THE FILE. IF WE DO THAT, THEN WE'LL BE ABLE TO COMPLETELY RE-
#		# CONTRUCT THE FlowSignals OBJECT FROM THE FLOW SIGNALS FILE.
#
#		# Try to open the file
#		try:
#			fid = tables.open_file(file_name,"r")
#		except IOError:
#			print "Could not open file %s" % file_name
#			sys.exit(-1)
#
#		# Read the time vector from the HDF file
#		self.time_vector = fid.root.time_vector.read()
#
#		# Loop over regions of interest (each region is a group under the group fid.root.sums)
#		for region_group in fid.root.sums:
#
#			# The name of the current region
#			region_name = region_group._v_title
#
#			# Create a dictionary entry under 'self.signals' for the current region
#			self.signals[region_name] = {}
#
#			# Loop over the signals for the current region of interest
#			for signal in fid.root.sums._f_getChild(region_name):
#
#				# The name of the current signal
#				signal_name = signal._v_name
#
#				# Add the current signal to the signals' dictionary
#				self.signals[region_name][signal_name] = signal.read()
#
#		# Close the file
#		fid.close()

# ----------------------------------------------------------------------------------------------- #

	def read_from_file(self, file_name):

		# Read the HDF file
		group_data = read_hdf_file(file_name)

		# Now, we try to re-construct the original FlowSignals object from the
		# information read from the HDF file. However, the HDF file doesn't have
		# all the information necessary to completely re-construct the original
		# FlowRegions object (flow_regions attribute of the FlowSignals object)

		# The time vector and the time signals
		self.time_vector = group_data['time_vector']
		self.signals     = group_data['sums']

		# The number of samples, sampling rate and initial time
		self.n_samples     = len(group_data['time_vector'])
		self.initial_time  = group_data['time_vector'][0]
		self.sampling_rate = 1/numpy.mean(numpy.diff(self.time_vector))

		# The names of the regions in the HDF file
		region_names = self.signals.keys()

		# Initialize the FlowRegions object
		self.flow_regions = FlowRegions(frame_size = (0, 0))

		# Now we fill the regions of interest in the FlowRegions object.
		# As the HDF file has no information about rods, no regions of
		# disinterest are filled. Also, since the regions' positions are
		# not saved either to the HDF file, we fill that with empty regions
		for region_name in region_names:
			# self.flow_regions.add_roi(region_name, region_position)
			self.flow_regions.add_roi(region_name, (0, 0, 0, 0))

# ----------------------------------------------------------------------------------------------- #

#	def write_to_xls_file(self,file_name):
#
#		# Create a work book
#		work_book = xlwt.Workbook()
#
#		# Add a sheet to the work book
#		flow_sheet = work_book.add_sheet('flow motion')
#
#		# The number of samples in the motion signals
#		# n_samples = len(time_vector)
#		n_samples = len(self.time_vector)
#
#		# The initial row in the Excel file where to start writing the motion signals.
#		# Here's row the first rows in the sheet are used:
#		# Row 0 -> signal name
#		# Row 1 -> sub-signal name
#		# Row 2 -> empty
#		initial_table_row = 3
#
#		# The number of the current column where to write data to
#		column_number = 0
#
#		# Write the name of the time vector signal to the Excel file
#		flow_sheet.write(0,column_number,"Time (s)")
#
#		# Loop over samples of the current signal
#		for m in range(n_samples):
#
#			# Write the current sample of the current signal to the output file
#			# flow_sheet.write(m+initial_table_row,column_number,time_vector[m])
#			flow_sheet.write(m+initial_table_row,column_number,self.time_vector[m])
#
#		# The signal names, in alphabetical order
#		# signal_names = signals.keys()
#		signal_names = self.signals.keys()
#		signal_names.sort()
#
#		# Loop over signals
#		for signal_name in signal_names:
#
#			# The sub-signal names, in alphabetical order
#			# sub_signal_names = signals[signal_name].keys()
#			sub_signal_names = self.signals[signal_name].keys()
#			sub_signal_names.sort()
#
#			# Loop over sub-signals ('x', 'y', 'mag', etc)
#			for sub_signal_name in sub_signal_names:
#
#				# Increment the column number
#				column_number = column_number+1
#
#				# Write the signal name to the current column in the Excel sheet
#				flow_sheet.write(0,column_number,signal_name)
#
#				# Write the sub-signal name to the current column in the Excel sheet
#				flow_sheet.write(1,column_number,sub_signal_name)
#
#				# The current signal
#				# signal = signals[signal_name][sub_signal_name]
#				signal = self.signals[signal_name][sub_signal_name]
#
#				# Number cells in Excel are kept as IEEE 754 64-bit floats (the same as the
#				# Python float type). The xlwt module detects common types (int, long, etc)
#				# and automatically convert them to float. For everything else, we have to
#				# do the conversion ourselves.
#				if (signal.dtype.type != numpy.float64):
#					signal = signal.astype('float64')
#
#				# Loop over samples of the current signal
#				for m in range(signal.size):
#
#					# Write the current sample of the current signal to the output file
#					flow_sheet.write(m+initial_table_row,column_number,signal[m])
#
#		# Save the file to disk
#		work_book.save(file_name)

# ----------------------------------------------------------------------------------------------- #

	def write_to_xls_file(self,file_name):

		# TODO: THE MODULE XLWT HAS A LIMIT OF 65536 ROWS PER SHEET. RIGHT NOW WE
		# DON'T CHECK FOR THAT BEFORE TRYING TO SAVE THE XLS FILE TO DISK. WE HAVE
		# TO FIX THAT. THE SIMPLEST SOLUTION IS TO JUST SKIP SAVING THE XLS FILE
		# IN CASE THERE ARE MORE THAN 65536 SAMPLES TO BE SAVED. A BETTER SOLUTION
		# WOULD BE TO CREATE AS MANY SHEETS AS ARE NECESSARY TO ACCOMMODATE THE
		# TOTAL NUMBER OF SAMPLES



		# There is a limit on the number of rows that an Excel sheet can have
		n_rows_limit = 65536

		# The initial row in the Excel file where to start writing the motion signals.
		# Here's row the first rows in the sheet are used:
		# Row 0 -> signal name
		# Row 1 -> sub-signal name
		# Row 2 -> empty
		initial_row = 3

		# The number of samples limit
		n_samples_limit = n_rows_limit - initial_row

		# The number of samples in the motion signals
		n_samples = len(self.time_vector)

		# The number of sheets needed to accommodate all the frames
		# n_sheets = int(numpy.ceil(float(n_samples)/float(n_rows_limit)))
		n_sheets = int(numpy.ceil(float(n_samples)/float(n_samples_limit)))

		# Create a work book
		work_book = xlwt.Workbook()

		# A list with the sheet objects
		flow_sheet = [None]*n_sheets

		# Add sheets to the work book
		for m in range(n_sheets):
			sheet_name = 'flow signals %g' % (m+1)
			flow_sheet[m] = work_book.add_sheet(sheet_name)

#		# The initial row in the Excel file where to start writing the motion signals.
#		# Here's row the first rows in the sheet are used:
#		# Row 0 -> signal name
#		# Row 1 -> sub-signal name
#		# Row 2 -> empty
#		initial_row = 3

		# The number of the current column where to write data to
		column_number = 0

		# Write the name of the time vector signal to the Excel file
		for m in range(n_sheets):
			flow_sheet[m].write(0,column_number,"Time (s)")

		# Loop over samples of the current signal
		for m in range(n_samples):


#			TODO: THIS IS WRONG. WE HAVE TO CORRECT IT. THE CORRECT
#			WAY IS PROBABLY SOMETHING LIKE:
#
#			sheet_number = m/(n_rows_limit-initial_row) + initial_row
#			row_number   = m%(n_rows_limit-initial_row) + initial_row

			# The sheet and row numbers
			# sheet_number = (m+initial_row)/n_rows_limit
			# row_number   = (m+initial_row)%n_rows_limit
			sheet_number = m/n_samples_limit
			row_number   = m%n_samples_limit + initial_row

			# Write the current sample of the time signal to the output file
			flow_sheet[sheet_number].write(row_number,column_number,self.time_vector[m])


		# The signal names, in alphabetical order
		signal_names = self.signals.keys()
		signal_names.sort()

		# Loop over signals
		for signal_name in signal_names:

			# The sub-signal names, in alphabetical order
			# sub_signal_names = signals[signal_name].keys()
			sub_signal_names = self.signals[signal_name].keys()
			sub_signal_names.sort()

			# Loop over sub-signals ('x', 'y', 'mag', etc)
			for sub_signal_name in sub_signal_names:

				# Increment the column number
				column_number = column_number+1

				# Write the signal and sub-signal names to the current column in the Excel sheet
				for m in range(n_sheets):
					flow_sheet[m].write(0,column_number,signal_name)
					flow_sheet[m].write(1,column_number,sub_signal_name)

				# The current signal
				# signal = signals[signal_name][sub_signal_name]
				signal = self.signals[signal_name][sub_signal_name]

				# Number cells in Excel are kept as IEEE 754 64-bit floats (the same as the
				# Python float type). The xlwt module detects common types (int, long, etc)
				# and automatically convert them to float. For everything else, we have to
				# do the conversion ourselves.
				if (signal.dtype.type != numpy.float64):
					signal = signal.astype('float64')

				# Loop over samples of the current signal
				for m in range(signal.size):

					# The sheet and row numbers
					# sheet_number = (m+initial_row)/n_rows_limit
					# row_number   = (m+initial_row)%n_rows_limit
					sheet_number = m/n_samples_limit
					row_number   = m%n_samples_limit + initial_row

					# Write the current sample of the current signal to the output file
					# flow_sheet.write(m+initial_row,column_number,signal[m])
					flow_sheet[sheet_number].write(row_number,column_number,signal[m])


		# Save the file to disk
		work_book.save(file_name)

# ----------------------------------------------------------------------------------------------- #

#	def write_to_pdf_file(self, file_name, signal_names = ('x','y','mag')):
#
#		# Plot the motion signals and export the plot to a pdf file
#		# figure_handle = self.plot_signals(('x','y','mag'))
#		figure_handle = self.plot_signals(signal_names, False)
#		figure_handle.savefig(file_name)

# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

#class FlowSignalsFile(object):
#
#
## ----------------------------------------------------------------------------------------------- #
#
#	def __init__(self, file_name, file_mode):
#
#
#		self.file_name    = file_name
#		self.file_mode    = file_mode
#		self.fid          = None
#		self.flow_regions = None
#
## ----------------------------------------------------------------------------------------------- #
#
#	def init_file(self, flow_source, flow_regions):
#
#		# Store the flow regions in the FlowSignalsFile object
#		self.flow_regions = flow_regions
#
#		# The names of the regions of interest
#		region_names = [region['name'] for region in self.flow_regions.rois]
#		region_names.sort()
#
#		# The signal names
#		signal_names = ['x', 'x_mag', 'y', 'y_mag', 'mag']
#		signal_names.sort()
#
#		# The number of frames to be stored in the output signals file
#		n_frames = flow_source.final_frame - flow_source.initial_frame
#
#		# The time vector. If we don't explicitly set the type to float here,
#		# we get a sequence of integers and, if the frame rate happens to be
#		# an integer as well, we end up with a sequence of integers for the
#		# time vector, which is wrong.
#		# self.time_vector = numpy.arange(self.initial_frame, self.final_frame, dtype=numpy.float)
#		# self.time_vector = self.time_vector/self.frame_rate
#		# time_vector = numpy.arange(self.initial_frame, self.final_frame, dtype=numpy.float)
#		time_vector = numpy.arange(flow_source.initial_frame, flow_source.final_frame, dtype=numpy.float)
#		time_vector = time_vector/flow_source.input_video.frame_rate
#
#		# Open the output file for writing
#		# fid = tables.open_file(file_name,"w")
#		self.fid = tables.open_file(self.file_name, self.file_mode)
#
#		# Create a group for storing the flow signals
#		# sums_group = fid.create_group("/", 'sums', 'Regions')
#		sums_group = self.fid.create_group("/", 'sums', 'Regions')
#
#		# The array shape and type in the HDF file
#		# array_shape = time_vector.shape
#		# array_type = time_vector.dtype
#		# array_shape = self.time_vector.shape
#		array_shape = time_vector.shape
#		# array_type = self.time_vector.dtype
#		array_type = time_vector.dtype
#		array_atom = tables.Atom.from_dtype(array_type)
#
#		# CArray filters
#		compression_method = "zlib"
#		#compression_method = "bzip2"
#		#compression_level = 0
#		compression_level = 2
#		array_filters = tables.Filters(complib=compression_method,complevel=compression_level)
#
#		# Create a CArray for storing the time vector
#		# array = fid.create_carray("/",'time_vector',array_atom,array_shape,filters=array_filters)
#		array = self.fid.create_carray("/",'time_vector',array_atom,array_shape,filters=array_filters)
#
#		# Write the time vector to a CArray
#		# array[:] = time_vector
#		# array[:] = self.time_vector
#		array[:] = time_vector
#
##		# The plot legends for all signals
##		legendas = {'mag': 'Magnitude','x_mag': 'Horz. Magnitude','y_mag': 'Vert. Magnitude','x': 'Horizontal','y': 'Vertical'}
#
#		# The number of regions
#		n_regions = len(region_names)
#
#		# The number of signals
#		n_signals = len(signal_names)
#
#		# The number of samples in the motion signals
#		# n_samples = signals[region_names[0]][signal_names[0]].shape[0]
#		# n_samples = len(time_vector)
#		# n_samples = len(self.time_vector)
#		n_samples = len(time_vector)
#
#		# Initialize a vector where to store the region groups
#		region_groups = [None]*n_regions
#
#		# Loop over regions, plotting the signals for each region in a separate panel (subplot)
#		for m in range(n_regions):
#
#			# Create a group in the HDF file for the current region
#			# region_groups[m] = fid.create_group(sums_group,region_names[m],region_names[m])
#			region_groups[m] = self.fid.create_group(sums_group,region_names[m],region_names[m])
#
#			# Loop over signals in the current region
#			for n in range(n_signals):
#
#				# Create an array in the HDF file for the current signal
#				# array = fid.create_carray(region_groups[m],signal_names[n],array_atom,array_shape,filters=array_filters)
#				# array = self.fid.create_carray(region_groups[m],signal_names[n],array_atom,array_shape,filters=array_filters)
#				self.fid.create_carray(region_groups[m],signal_names[n],array_atom,array_shape,filters=array_filters)
#
##				# Write the current signal to the current CArray
##				# array[:] = signals[region_names[m]][signal_names[n]]
##				array[:] = self.signals[region_names[m]][signal_names[n]]
#
##		# Close the file
##		fid.close()
#
#
## ----------------------------------------------------------------------------------------------- #
#
#	def close_file(self):
#
#		# Close the optical flow file
#
#		# If the file identifier 'fid' is an empty matrix, the
#		# file is already closed and then we have nothing to do
#		if self.fid is None:
#			print "File %s is already closed." % self.file_name
#			return
#
#		# Try to close the optical flow file. If it succeeds,
#		# we assign an empty matrix to the file identifier
#		# to signal that the file is closed
#		try:
#			self.fid.close()
#			self.fid = None
#		except IOError:
#			print "Could not close file %s" % self.file_name
#			sys.exit(-1)
#
## ----------------------------------------------------------------------------------------------- #
#
#
#	def process_rods(self,flow_frame):
#
#		# Loop over regions
#		# for region in self.rods:
#		for region in self.flow_regions.rods:
#
#			# The position of the current region
#			position = region['position']
#
#			# Loop over the flow directions (x and y)
#			for key in flow_frame.keys():
#				flow_frame[key][position[1]:position[3],position[0]:position[2]] = 0.0
#
## ----------------------------------------------------------------------------------------------- #
#
#	def process_rois(self,flow_frame,frame_number):
#
#		# IMPORTANT: REGION COORDINATES ARE SPECIFIED AS [X1, Y1, X2, Y2]
#
#		# Add magnitude channels to the dictionary 'self.flow_frame'
#		flow_frame['x_mag'] = numpy.absolute(flow_frame['x'])
#		flow_frame['y_mag'] = numpy.absolute(flow_frame['y'])
#		flow_frame['mag']   = numpy.sqrt(numpy.square(flow_frame['x']) + numpy.square(flow_frame['y']))
#
#		# Loop over regions
#		# for region in self.rois:
#		for region in self.flow_regions.rois:
#
#			# The name of the current position
#			region_name = region['name']
#
#			# The position of the current region
#			position = region['position']
#
#			# The number of pixels in the current region of interest (for normalization purposes)
#			# n_pixels = (position[3]-position[1]+1)*(position[2]-position[0]+1)
#			n_pixels = (position[3]-position[1])*(position[2]-position[0])
#
#			# Loop over the flow signals
#			for signal_name in flow_frame.keys():
#				# sums[region['name']][key] = numpy.sum(flow[key][position[1]:position[3],position[0]:position[2]])
#				# self.signals[region_name][signal_name][frame_number] = numpy.sum(flow_frame[signal_name][position[1]:position[3],position[0]:position[2]])/n_pixels
#				self.fid.root.sums._f_getChild(region_name)._f_getChild(signal_name)[frame_number] = \
#				numpy.sum(flow_frame[signal_name][position[1]:position[3],position[0]:position[2]])/n_pixels
#
#
#
#
##
##
##			# Loop over signals in the current region
##			for n in range(n_signals):
##
##				# Create an array in the HDF file for the current signal
##				# array = fid.create_carray(region_groups[m],signal_names[n],array_atom,array_shape,filters=array_filters)
##				# array = self.fid.create_carray(region_groups[m],signal_names[n],array_atom,array_shape,filters=array_filters)
##				self.fid.create_carray(region_groups[m],signal_names[n],array_atom,array_shape,filters=array_filters)
##
###				# Write the current signal to the current CArray
###				# array[:] = signals[region_names[m]][signal_names[n]]
###				array[:] = self.signals[region_names[m]][signal_names[n]]
##
###		# Close the file
###		fid.close()
##
##
#
#
#
## ----------------------------------------------------------------------------------------------- #
#
#	# def process_flow_frame(self, flow_frame, frame_number):
#	def process_flow_frame(self, frame, frame_number):
#
#
#		# IMPORTANT: We have a problem here. If the flow frame is coming from the
#		# FlowAnalysis class, it comes as an OpenCV image (iplimage), whereas if
#		# it comes from the FlowReader class, it comes as a numpy array (the
#		# 'read_frame' method of both FlowFileBinary and FlowFileHDF classes
#		# converts the OpenCV image to a numpy array using numpy.fromstring().
#		# As a quick workaround, we check for that here and convert the flow
#		# frame to a numpy array in case it is in the OpenCV format. In the
#		# long term, I think the best solution is to drop the use of OpenCV
#		# images and work only with numpy arrays (however, I'm not totally
#		# sure that's possible; I know OpenCV these days tend to favor the
#		# use of CVMat images instead of IplImages, but I'm not sure the
#		# CVMat is a numpy format).
#
#
##		# HERE'S HOW IT IS DONE WHEN READING A BIN FILE
##		# Read the x and y flow frames from the binary file
##		flow['x'] = numpy.fromstring(self.fid.read(n_bytes),self.dtype_string).reshape(dims).astype("float64")
##		flow['y'] = numpy.fromstring(self.fid.read(n_bytes),self.dtype_string).reshape(dims).astype("float64")
##
##		# HERE'S HOW IT IS DONE WHEN READING AN HDF FILE
##		# Read the flow values from the CArray. Interestingly, this works not only for strings,
##		# but for Numpy arrays as well. Weird or what?
##		flow['x'] = numpy.fromstring(self.fid.root.flow.x[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float64")
##		flow['y'] = numpy.fromstring(self.fid.root.flow.y[frame_number][0:n_bytes],self.dtype_string).reshape(dims).astype("float64")
#
#		# cv.iplimage
#		# cv.cvmat
#
#
#		# We need to create a new dictionary variable 'flow_frame' and then copy the
#		# contents of 'frame' to the new variable in order to avoid changing the
#		# contents of the original variable. If we don't do that, the contents of
#		# the original variable 'frame' will change even in the scope of the calling
#		# function, and we definetily don't want that, especially because we do a
#		# conversion from IplImage to cvMat here
#		flow_frame = {}
#
#		# If the flow frames are IplImages, we convert them to Numpy arrays. If the
#		# flow frames are coming from a FlowReader object which is reading the flow
#		# information from file, the flow frames will be Numpy arrays as both the
#		# FlowFileBinary and FlowFileHDF classes do the conversion upon reading the
#		# information from disk. On the other hand, if the flow frames are coming
#		# from the FlowAnalysis class, they will be IplImages and therefore we have
#		# to do the conversion here.
#		# TODO: I think the best long term solution is to work with cvMat arrays
#		# instead of IplImages throughout our code. However, in order to do that,
#		# we need to update the writing and reading of flow frames to/from disk.
#		# if (type(flow_frame["x"]) == cv.iplimage):
#		if (type(frame["x"]) == cv.iplimage):
#
#			# flow_depth = flow_frame["x"].depth
#			flow_depth = frame["x"].depth
#			dtype_string = FlowFileBinary.get_dtype_string(flow_depth)
#
#			# The frame dimensions
#			# dims = (flow_frame["x"].height, flow_frame["x"].width)
#			dims = (frame["x"].height, frame["x"].width)
#
#
#		# flow_frame_x = numpy.fromstring(flow_frame["x"].tostring(),dtype_string).reshape(dims).astype("float64")
#		# flow_frame_y = numpy.fromstring(flow_frame["y"].tostring(),dtype_string).reshape(dims).astype("float64")
#		# flow_frame["x"] = numpy.fromstring(flow_frame["x"].tostring(),dtype_string).reshape(dims).astype("float64")
#		# flow_frame["y"] = numpy.fromstring(flow_frame["y"].tostring(),dtype_string).reshape(dims).astype("float64")
#		flow_frame["x"] = numpy.fromstring(frame["x"].tostring(),dtype_string).reshape(dims).astype("float64")
#		flow_frame["y"] = numpy.fromstring(frame["y"].tostring(),dtype_string).reshape(dims).astype("float64")
#
#
#		# Process the regions of disinterest. Important: these regions must
#		# be processed *before* the regions of interest
#		# if self.rods:
#		if self.flow_regions.rods:
#			self.process_rods(flow_frame)
#
#		# Process the regions of disinterest.
#		# if self.rois:
#		if self.flow_regions.rois:
#			self.process_rois(flow_frame,frame_number)
#

# ----------------------------------------------------------------------------------------------- #
# ----------------------------------------------------------------------------------------------- #

class FlowReader(object):


	# This class is the reader equivalent of the FlowAnalysis class. Basically, it takes an
	# input flow file and a frame range and process that range. From each frame read, it
	# should emit a 'frame_read' signal that can then be connected to some method of an
	# external GUI, for example.


# ----------------------------------------------------------------------------------------------- #

	#----- Signals emitted by this class -----#

	# Signal emitted when the flow reading is canceled
	# flow_reading_canceled = notify.Signal()

	# Signal emitted after an optical flow frame has been read
	# flow_frame_read = notify.Signal()
	flow_frame_processed = notify.Signal()

	# Signal emitted when the flow reading is finished
	# flow_reading_finished = notify.Signal()
	flow_source_finished = notify.Signal()

# ----------------------------------------------------------------------------------------------- #

	# def __init__(self, input_file_name, regions = None):
	def __init__(self, input_file_name):

		# The name of the input file
		self.input_file_name = input_file_name

		# The input flow file
		self.input_flow_file = None

		# The regions of interest/disinterest
		# self.regions = regions

		# The resolution of the input video
		self.width  = None
		self.height = None

		# The frame rate of the input video
		self.frame_rate = None

		# The initial and final frames for the analysis
		self.initial_frame = None
		self.final_frame   = None

		# The first and last flow frames in the input flow file
		self.limit_frames = None

		# The number of the current optical flow frame
		self.frame_number = -1

		# This attribute is used to cancel the flow reading process
		self.flow_reading_canceled = False

		# Variables holding optical flow frames
		self.flow_frame = None

		# Open the input file
		self.open_input_file()

# ----------------------------------------------------------------------------------------------- #

	def open_input_file(self):

		# The extension of the input file
		(zzz,extension) = os.path.splitext(self.input_file_name)

		# Valid file extensions
		valid_extensions = [file_type['file_extension'] for file_type in FlowApplication.flow_file_types]

		# Check whether the input file has a valid extension
		if extension.lower() not in valid_extensions:
			error_message = "%s: Invalid file type" % self.input_file_name
			raise Exception(error_message)

		# The flow class to be used to read the input file
		(flow_class,) = [file_type['function'] for file_type in FlowApplication.flow_file_types if file_type['file_extension'] == extension]

		# Create a flow file object
		self.input_flow_file = flow_class(self.input_file_name,"r",None)

		# Open the flow file
		self.input_flow_file.open_file()

		# Read the file header. This must be done before reading any flow frames
		self.input_flow_file.read_file_header()

		# Set the input video resolution
		(self.width, self.height) = (self.input_flow_file.file_header['n_cols'], self.input_flow_file.file_header['n_rows'])

		# Set the input video's frame rate
		self.frame_rate = self.input_flow_file.file_header['frame_rate']

		# The first and last flow frames in the input flow file
		# self.limit_frames = (self.input_flow_file.file_header['initial_frame'], self.input_flow_file.file_header['final_frame']-1)
		self.limit_frames = (self.input_flow_file.file_header['initial_frame'], self.input_flow_file.file_header['final_frame'])

		# Set the frame range to its default value (i.e., to the first
		# and last frames in the input video). The frame range can be
		# changed later by calling the 'set_frame_range' method again
		self.set_frame_range()

# ----------------------------------------------------------------------------------------------- #

#	def set_default_region(self):
#
#		# The height and width of the input video/flow frame
#		height = self.input_flow_file.file_header['n_rows']
#		width  = self.input_flow_file.file_header['n_cols']
#
#		# A dictionary with the current region's properties
#		region = {}
#		region["region_widget"] = None
#		region["text_widget"] = None
#		region["name"] = "frame"
#		region["type"] = "interest"
#		region["position"] = (0, 0, width, height)
#
#		# Append the new region to the list of regions in the FlowReader object
#		self.regions.append(region)
#
#		# Return the created region
#		return region

# ----------------------------------------------------------------------------------------------- #

#	def set_regions(regions = None):
#
#		pass

# ----------------------------------------------------------------------------------------------- #

	def set_frame_range(self, initial_frame = None, final_frame = None):

		# If the initial frame has not been provided, it defaults to the first frame in the flow file
		if initial_frame is None:
			initial_frame = self.limit_frames[0]

		# If the final frame has not been provided, it defaults to the last frame in the flow file
		if final_frame is None:
			final_frame = self.limit_frames[1]

		# Check if the initial frame is within the valid frame range
		if not (self.limit_frames[0] <= initial_frame <= self.limit_frames[1]):
			error_message = "Invalid initial frame. Frame range is [%g %g]." % (self.limit_frames[0], self.limit_frames[1])
			raise Exception(error_message)

		# Check if the final frame is within the valid frame range
		if not (self.limit_frames[0] <= final_frame <= self.limit_frames[1]):
			error_message = "Invalid final frame. Frame range is [%g %g]." % (self.limit_frames[0], self.limit_frames[1])
			raise Exception(error_message)

		# The final frame must be greater than the initial frame
		if (initial_frame >= final_frame):
			raise Exception("The initial frame must be smaller than the final frame.")

		# Update the frame range attributes ('initial_frame', 'final_frame', 'n_frames')
		# in the FlowAnalysis object
		self.initial_frame = initial_frame
		self.final_frame   = final_frame

# ----------------------------------------------------------------------------------------------- #

#	def update_analysis_status(self,status):
#
#		# This method handles the signal 'analysis_canceled' emitted by the class
#		# FlowGui when the flow analysis is canceled through the progress bar.
#
#		# self.flow_analysis_canceled = status
#		self.processing_canceled = status

# ----------------------------------------------------------------------------------------------- #

	def on_canceled_processing(self):

		# This method handles the signal 'analysis_canceled' emitted by the class
		# FlowGui when the flow analysis is canceled through the progress bar.

		# self.flow_analysis_canceled = status
		self.flow_analysis_canceled = True

# ----------------------------------------------------------------------------------------------- #


	def read_frame(self,frame_number):

		# Read the frame number of interest
		self.input_flow_file.read_frame(frame_number)

		# Update the 'flow_frame' attribute in the FlowReader object
		self.flow_frame = self.input_flow_file.flow_frame

		# Convert the read frame from a numpy array to an OpenCV image
		# TODO: IN THE FUTURE, I THINK WE SHOULD USE NUMPY ARRAYS TO STORE IMAGES
		# REGARDLESS OF THE VIDEO BACKEND WE'RE USING (EITHER OPENCV OR PYFFMPEG)
		self.flow_frame = {}
		self.flow_frame['x'] = cv.GetImage(cv.fromarray(self.input_flow_file.flow_frame['x']))
		self.flow_frame['y'] = cv.GetImage(cv.fromarray(self.input_flow_file.flow_frame['y']))

		# Update the 'frame_number' attribute in the FlowReader object
		self.frame_number = self.input_flow_file.frame_number

		# Emit the 'flow_frame_processed' signal
		# self.flow_frame_read(self.flow_frame,self.frame_number)
		self.flow_frame_processed(self.flow_frame,self.frame_number)

# ----------------------------------------------------------------------------------------------- #

	def read_next_frame(self):

		# Read the frame number of interest
		self.input_flow_file.read_next_frame()

		# Update the 'flow_frame' attribute in the FlowReader object
		self.flow_frame = self.input_flow_file.flow_frame

		# Convert the read frame from a numpy array to an OpenCV image
		# TODO: IN THE FUTURE, I THINK WE SHOULD USE NUMPY ARRAYS TO STORE IMAGES
		# REGARDLESS OF THE VIDEO BACKEND WE'RE USING (EITHER OPENCV OR PYFFMPEG)
		self.flow_frame = {}
		self.flow_frame['x'] = cv.GetImage(cv.fromarray(self.input_flow_file.flow_frame['x']))
		self.flow_frame['y'] = cv.GetImage(cv.fromarray(self.input_flow_file.flow_frame['y']))

		# Update the 'frame_number' attribute in the FlowReader object
		self.frame_number = self.input_flow_file.frame_number

		# Emit the 'flow_frame_processed' signal
		# self.flow_frame_read(self.flow_frame,self.frame_number)
		self.flow_frame_processed(self.flow_frame,self.frame_number)

# ----------------------------------------------------------------------------------------------- #

	def run(self):


#		# NOTE: THIS METHOD STARTED OUT AS A SIMPLE COPY OF THE METHOD
#		# read_flow_file() IN THE FILE optical_flow_cv21_oo.py
#		# TODO: I THINK IT WOULD BE A GOOD IDEA TO ADD THIS METHOD TO THE
#		# MODULE OPENCV_LIB_4_CV21. THIS WOULD BE A METHOD THAT WOULD TAKE
#		# THE REGIONS OF INTEREST AND/OR DISINTEREST AND RETURN ALL THE
#		# SIGNALS COMPUTED FOR EACH OF THE REGIONS.
#
#		# IMPORTANT!!! POSITION SPECIFIED AS [XI YI XF YF]. VERY IMPORTANT: THE UPPER
#		# LIMIT IS NOT INCLUDED (THAT'S HOW PYTHON WORKS!!!)
#
##		# This region was obtained by loading the file 'CL1JD1Test_test_regions.mat'
##		# roi_1 = {'type': 'roi', 'name': 'face', 'position': [242, 29, 600, 362]}
##		# roi_1 = {'type': 'roi', 'name': 'face', 'position': [241, 28, 599, 361]}
##		roi_1 = {'type': 'roi', 'name': 'frame', 'position': [0, 0, 720, 480]}
##
##		# rods = [rod_1, rod_1, rod_1]
##		rods = numpy.array([])
##		# rois = [roi_1, roi_2, roi_3]
##		rois = [roi_1]
#
#
#
#		# If 'self.regions' is empty, we assign it a default region of interest
#		# encompassing the whole picture frame
#		if (self.regions is None):
#			self.regions = []
#
#		# The regions of interest and disinterest
#		rois = [region for region in self.regions if (region["type"].lower() == "interest")]
#		rods = [region for region in self.regions if (region["type"].lower() == "disinterest")]
#
#		# TODO: WHEN NO REGIONS OF INTEREST HAVE BEEN DEFINED, WE HAVE TO PRESENT
#		# THE USER WITH A POPUP WINDOW LETTING HIM KNOW THAT WE'RE GOING TO CREATE
#		# A DEFAULT ROI ENCOMPASSING THE WHOLE SCENE
#
#		# If no regions of interest have been defined, we define a default one
#		# containing the whole frame and append it to the list of regions in
#		# the FlowReader object
#		if not rois:
#			region = self.set_default_region()
#
#			# Append the new region of the list of regions of interest (rois)
#			rois.append(region)
#
#
#
##		# The initial and final frames, gotten from the spinboxes. Please,
##		# note that these values may be different from those in the optical
##		# flow file. The values in the optical flow file tells us the range
##		# of *video* frames used in computing the optical flow file. The
##		# values in the spinboxes tells us the range of *optical flow*
##		# frames to be analyzed
##		initial_frame = self.initial_frame_spinbox.value()
##		final_frame = self.final_frame_spinbox.value()
#
#
#
#		# The number of optical flow frames to be analyzed
#		# n_frames = final_frame - initial_frame
#		n_frames = self.final_frame - self.initial_frame + 1
#
#		# Initialize the variable 'sums', where the summed flow within each region
#		# of interest will be stored
#		sums = [None]*n_frames
#
#
##		# Define a progress bar
##		progress_bar = QtGui.QProgressDialog(self)
##		progress_bar.setMinimumDuration(2000)
##		progress_bar.setLabelText("Processing frames...")
##		progress_bar.setRange(initial_frame,final_frame)
##		progress_bar.setWindowModality(QtCore.Qt.WindowModal)
#
#
##		# Just a shortcut to the flow file object
##		flow_file = self.input_flow_file
#
##		# Start our stopwatch
##		start_time = time.time()



		# self.input_flow_file.read_frame(self.initial_frame)
		self.read_frame(self.initial_frame)


		# Read all flow frames in the bin file
		# for frame_number in range(initial_frame,final_frame):
		while (self.frame_number < (self.final_frame-1)):

#			indice = frame_number - initial_frame
#
#
#			progress_bar.setValue(frame_number)
#
#			if (progress_bar.wasCanceled()):
#				# break
#				return




			self.read_next_frame()
			print "Reading frame %g" % self.frame_number



#			# flow_file.read_next_frame()
#			# flow_file.read_frame(frame_number)
#			flow_file.read_frame(indice)
#			flow_file.remove_rods(rods)
#			# sums[m] = flow_file.process_rois(rois)
#			sums[indice] = flow_file.process_rois(rois)
#			# print "Reading frame %g" % flow_file.frame_number
#			print "Reading frame %g" % frame_number




		# Emit the 'flow_source_finished' signal to indicate that the
		# flow reading is finished.
		# self.flow_reading_finished()
		self.flow_source_finished()





		return

#
#		# TODO: all the code below should go into the class FlowSignals.
#
#
##		# Finalize the progress bar and close it
##		progress_bar.setValue(final_frame)
#
#		# Re-organize the contents of the variable 'sums' into the variable 'signals'
#		signals = self.__rearrange_motion_signals(sums,initial_frame,final_frame)
#
#		# Stop our stopwatch and print out the elapsed time
#		elapsed_time = (time.time() - start_time)
#		print "Elapsed time = %g " % elapsed_time
#
#		# The file names where to save the motion signals
#		(base_file_name,extension) = os.path.splitext(flow_file.file_name)
#		pdf_file_name   = "%s_motion_signals.pdf" % base_file_name
#		hdf_file_name   = "%s_motion_signals.hdf" % base_file_name
#		excel_file_name = "%s_motion_signals.xls" % base_file_name
#
#		# The time vector. If we don't explicitly set the type to float here,
#		# we get a sequence of integers and, if the frame rate happens to be
#		# an integer as well, we end up with a sequence of integers for the
#		# time vector, which is wrong.
#		frame_rate  = self.input_flow_file.file_header['frame_rate']
#		time_vector = numpy.arange(initial_frame,final_frame,dtype=numpy.float)
#		time_vector = time_vector/frame_rate
#
#		# Plot the motion signals
#		# self.__plot_motion_signals(time_vector,signals)
#		# self.__plot_motion_signals(time_vector,signals,('x','y','mag'))
#		figure_handle = self.__plot_motion_signals(time_vector,signals,('x','y','mag'))
#		figure_handle.savefig(pdf_file_name)
#
#		# Save motion signals to HDF file
#		# hdf_file_name = "motion_signals.hdf"
#		self.__write_flow_signals_to_hdf_file(time_vector,signals,hdf_file_name)
#
#		# Save motion signals to Excel file
#		# excel_file_name = "motion_signals.xls"
#		self.__write_excel_file(time_vector,signals,excel_file_name)
#
#		# OK, we're done
#		print "Motion signals have bee saved to the files %s.(hdf, xls, pdf)" % base_file_name

# ----------------------------------------------------------------------------------------------- #


def read_hdf_file(file_name):

# --------------------------------------------------------------------------- #

	def read_group(group):

		# Initialize the group data dictionary
		group_data = {}

		# List of groups under the current group
		group_names = group._v_groups

		# List of leaves (data sets) under the current group
		leaf_names = group._v_leaves

		# Loop over leaves
		for leaf_name in leaf_names:
			group_data[leaf_name] = group._f_getChild(leaf_name).read()

		# Loop over groups
		for group_name in group_names:
			group_data[group_name] = read_group(group._f_getChild(group_name))

		# Return the group data
		return group_data

# --------------------------------------------------------------------------- #

	# Try to open the input file for reading
	try:
		fid = tables.open_file(file_name,"r")
	except IOError:
		print "Could not open file %s" % file_name
		sys.exit(-1)

	# Recursively read everything under the root group
	group_data = read_group(fid.root)

	# Close the file
	fid.close()

	# Return the data read from file
	return group_data

# ----------------------------------------------------------------------------------------------- #
