#!/usr/bin/python
# -*- coding: utf-8 -*-

# Author: Adriano Vilela Barbosa

# ----------------------------------------------------------------------------------------------- #

import matplotlib.pyplot as plt
import numpy as np
import flow_lib
import cv2
import os

# ----------------------------------------------------------------------------------------------- #

def perform_flow_analysis(input_file_name, regions_file_name = None, output_file_name = None,
	initial_frame = None, final_frame = None):

	# Create a MovieReaderOpencv object for handling the input video.
	# The only reason we need to do this here is to find the video's
	# resolution (frame size) in order to pass it to the function
	# responsible for creating the regions of interest/disinterest
	input_video = flow_lib.MovieReaderOpencv(input_file_name)

	# The video's frame size
	# frame_size = (720, 480)
	frame_size = (input_video.width, input_video.height)

	# Define the regions of interest and/or disinterest
	if regions_file_name is None:

		# If no regions file has been provided, we make flow_regions = []
		# to signal that we want a single region of interest encompassing
		# the whole frame
		flow_regions = []

	else:

		# Create a FlowRegions object
		flow_regions = flow_lib.FlowRegions(frame_size)

		# Read regions of interest/disinterest from the regions file
		flow_regions.load_from_file(regions_file_name)

	# Construct the FlowApplication object
	flow_app = flow_lib.FlowApplication(
		input_file_name,
		flow_regions,
		output_file_name,
		initial_frame,
		final_frame)

	# Run the flow application
	flow_app.run()

# ----------------------------------------------------------------------------------------------- #

def linear_mapping(x,L=255.0,xmin=0,xmax=1):

	# Computes a linear function given by
	#
	# f(x) = L*(x-xmin)/(xmax-xmin)

	# This is the equation of a line passing through the points
	# (xmin,0) and (xmax,L)

	# Flow values below xmin are mapped to zero
	# Flow values above xmax are mapped to L
	# Flow values between xmin and xmax are linearly mapped to the range [0, L]

	xmin = L*xmin
	xmax = L*xmax

	x[x < xmin] = xmin;
	x[x > xmax] = xmax;

	return L*(x-xmin)/(xmax-xmin)

# ----------------------------------------------------------------------------------------------- #

def logistic_mapping(x,L=1,k=1,x0=0):

	# Computes a logistic mapping given by
	#
	# f(x) = L/(1 + exp(-k*(x-x0)))
	#
	# where
	#
	# L:  the curve's maximum value
	# k:  the steepness of the curve
	# x0: the sigmoid's midpoint along the x axis

	return L/(1+np.exp(-k*(x-x0)))

# ----------------------------------------------------------------------------------------------- #

def create_flow_movie(flow_field_file_name,flow_video_file_name=None,xmin=0,xmax=1):

	# A boolean to choose between color and grayscale video.
	# Note: the flow video is always a grayscale video. However, it is possible
	# to make a grayscale color video by making all the three color channels
	# the same. Why would anyone want to do that? Well, at least for the XVID
	# codec, the latter approach results in videos which are about 25% smaller.
	# This is counter-intuitive, and I don't know why it happens. I still have
	# to test this with other video codecs though
	color_video = True

	# The codec used to encode the output video. Note: apparently, on a Mac,
	# we need to use 'mp4v', in lower case. See this link for more details:
	# http://stackoverflow.com/questions/10605163/opencv-videowriter-under-osx-producing-no-output
	# fourcc = cv2.cv.CV_FOURCC(*'X264')
	# fourcc = cv2.cv.CV_FOURCC(*'MJPG')
	# fourcc = cv2.cv.CV_FOURCC(*'XVID')
	fourcc = cv2.cv.CV_FOURCC(*'mp4v')

	# If a name has not been provided for the output flow video, we use a default one
	if flow_video_file_name is None:
		aux = os.path.splitext(flow_field_file_name)
		flow_video_file_name = aux[0] + '.mov'

	# Instantiate a FlowFileHDF object for reading the flow field file
	flow_file = flow_lib.FlowFileHDF(flow_field_file_name,"r",None)

	# Open the flow file
	flow_file.open_file()

	# Read the file header. This must be done before reading any flow frames
	flow_file.read_file_header()

	# The input video file used in creating the optical flow file
	# input_video_file_name = flow_file.file_header['input_video_file']

	# The frame rate of the original video. We'll use this frame
	# rate for the flow video as well
	frame_rate = flow_file.file_header['frame_rate']

	# The number of flow frames in the flow field file
	n_frames = flow_file.n_frames
	# n_frames = 200

	# The frame dimensions
	n_rows = flow_file.file_header['n_rows']
	n_cols = flow_file.file_header['n_cols']
	n_channels = flow_file.file_header['n_channels']

	# Initialize the output video
	video = cv2.VideoWriter(flow_video_file_name,fourcc,frame_rate,(n_cols,n_rows),color_video)

	# Loop over all flow frames
	while (flow_file.frame_number < (n_frames-1)):

		try:

			# Read the next flow frame
			flow_file.read_next_frame()

		except:

			break

		# Print out a progress message
		print 'Processing frame %g/%g' % (flow_file.frame_number+1,n_frames)

		# Compute the flow magnitude for the current frame
		flow_x = flow_file.flow_frame['x']
		flow_y = flow_file.flow_frame['y']
		flow_mag = np.sqrt(flow_x**2 + flow_y**2)

		# Use a logistic function to map the magnitude values to the range [0, 255]
		# flow_mag = logistic_mapping(flow_mag,L=255,k=0.05,x0=100)
		# flow_mag = logistic_mapping(flow_mag,L=255,k=0.04,x0=100)
		flow_mag = linear_mapping(flow_mag,xmin=xmin,xmax=xmax)

		# Convert 'flow_mag' from float32 to uint8, in order to write it
		# as a video frame
		flow_mag = flow_mag.astype('uint8')

		# If 'color_video' is true, we create a grayscale color video by making
		# all three color channels the same
		if color_video:
			frame = np.tile(flow_mag[:,:,np.newaxis],[1,1,3])
		else:
			frame = flow_mag

		# Add current frame to the video
		video.write(frame)

	video.release()
	cv2.destroyAllWindows()

# ----------------------------------------------------------------------------------------------- #

def create_video_from_images(input_dir,output_file_name,frame_rate = 30):

	# The codec used to encode the output video
	# fourcc = cv2.cv.CV_FOURCC(*'X264')
	# fourcc = cv2.cv.CV_FOURCC(*'MJPG')
	# fourcc = cv2.cv.CV_FOURCC(*'XVID')
	fourcc = cv2.cv.CV_FOURCC(*'mp4v')

	frame_rate = float(frame_rate)

	# The list of files in the directory 'input_dir'. This does NOT include directory names
	file_list = [os.path.normcase(f) for f in os.listdir(input_dir) if os.path.isfile(os.path.join(input_dir,f))]

	# For some reason, the file names in the list are not sorted. Thus, we call the
	# sort() method in order to sort the file names in alphabetical order
	file_list.sort()

	# The full path to the first image of the sequence
	file_name = os.path.join(input_dir,file_list[0])

	# Read the first image of the sequence
	img = cv2.imread(file_name)

	# The image dimensions
	(height, width, layers) = img.shape

	# Initialize the output video
	# video = cv2.VideoWriter('video.avi',-1,1,(width,height))
	# video = cv2.VideoWriter('video.avi',fourcc,20.0,(width,height))
	video = cv2.VideoWriter(output_file_name,fourcc,frame_rate,(width,height))

	for file_name in file_list:

		# The full path to the current file
		file_name = os.path.join(input_dir,file_name)

		# Read the image in the current file
		img = cv2.imread(file_name)

		# Add current frame to the video
		video.write(img)

	cv2.destroyAllWindows()
	video.release()

# ----------------------------------------------------------------------------------------------- #

def plot_flow_stats(flow_field_file_name):

	# Instantiate a FlowFileHDF object for reading the flow field file
	flow_file = flow_lib.FlowFileHDF(flow_field_file_name,"r",None)

	# Open the flow file
	flow_file.open_file()

	# Read the file header. This must be done before reading any flow frames
	flow_file.read_file_header()

	# The input video file used in creating the optical flow file
	# input_video_file_name = flow_file.file_header['input_video_file']

	# The number of flow frames in the flow field file
	n_frames = flow_file.n_frames
	# n_frames = 10

	# The frame dimensions
	n_rows = flow_file.file_header['n_rows']
	n_cols = flow_file.file_header['n_cols']
	n_channels = flow_file.file_header['n_channels']

	# Initialize 'flow_mag', a matrix for holding the magnitude values
	# of an optical flow frame. We need it here in order to initialize
	# the flow magnitude histogram
	flow_mag = np.zeros((n_rows,n_cols),dtype='float32')

	# The number of bins and the flow range to be used in computing
	# the flow magnitude histogram
	n_bins = 1000
	# hist_range = [0,655]
	hist_range = [10,400]

	# Initialize the flow magnitude histogram
	hist_mag = cv2.calcHist([flow_mag],[0],None,[n_bins],hist_range)

	# A vector for holding the maximum flow magnitude value of each frame
	max_vector = np.zeros(n_frames)

	# Loop over all flow frames
	for k in range(n_frames):

		try:

			# Read the next flow frame
			flow_file.read_next_frame()

		except:

			break

		# Print out a progress message
		print 'Processing frame %g/%g' % (flow_file.frame_number+1,n_frames)

		# Compute the flow magnitude for the current frame
		flow_x = flow_file.flow_frame['x']
		flow_y = flow_file.flow_frame['y']
		flow_mag = np.sqrt(flow_x**2 + flow_y**2)

		# Update the flow magnitude histogram with the contents of the current
		# flow magnitude frame
		hist_mag = cv2.calcHist([flow_mag],[0],None,[n_bins],hist_range,hist_mag,True)

		# The maximum flow magnitude for the current frame
		max_vector[k] = flow_mag.max()

	# The CDF (Cumulative Distribution Function) of the flow mag values
	cdf_mag = np.cumsum(hist_mag)/np.sum(hist_mag)

	# Plot the logistic function
	x = np.arange(0,655,1)
	y = logistic_mapping(x,k=0.025,x0=655.0/2)
	# y = logistic_mapping(x,k=0.01,x0=655.0/2)
	fig_1, ax_1 = plt.subplots()
	ax_1.plot(x,y)
	ax_1.set_title('Logistic function')

	# Plot the flow magnitude histogram
	x_vector = np.linspace(hist_range[0],hist_range[1],n_bins)
	bar_width = 0.8*np.diff(hist_range)[0]/n_bins
	fig_2, ax_2 = plt.subplots()
	ax_2.bar(x_vector,hist_mag,bar_width)
	ax_2.set_title('Distribution of flow magnitude')
	ax_2.set_xlabel('Flow magnitude')
	ax_2.set_ylabel('Number of hits')
	# fig_2.savefig('flow_mag_histogram.eps')
	fig_2.savefig('flow_mag_histogram.pdf')

	# Plot the flow magnitude histogram
	# x_vector = np.linspace(hist_range[0],hist_range[1],n_bins)
	# bar_width = 0.8*np.diff(hist_range)[0]/n_bins
	fig_3, ax_3 = plt.subplots()
	ax_3.plot(x_vector,cdf_mag,'.-')
	ax_3.set_title('CDF of flow magnitude')
	ax_3.set_xlabel('Flow magnitude')
	ax_3.set_ylabel('Accumulated hits (%)')
	ax_3.grid('on')
	# fig_3.savefig('flow_mag_cdf.eps')
	fig_3.savefig('flow_mag_cdf.pdf')

#	# Plot the maximum flow magnitudes
#	fig_4, ax_4 = plt.subplots()
#	ax_4.plot(max_vector, "o")
#	ax_4.set_title('Maximum magnitude flow value per frame')
#	ax_4.set_xlabel('Frame number')
#	ax_4.set_ylabel('Maximum magnitude value')
#	# fig_4.savefig('flow_mag_max_values.eps')
#	fig_4.savefig('flow_mag_max_values.pdf')

	plt.show()

# ----------------------------------------------------------------------------------------------- #
